"""
******************************************************************************

    Echoes Data Browser (Ebrow) is a data navigation and report generation
    tool for Echoes.
    Echoes is a RF spectrograph for SDR devices designed for meteor scatter
    Both copyright (C) 2018-2023
    Giuseppe Massimo Bertani gm_bertani(a)yahoo.it

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html

*******************************************************************************

"""
import os
import sys
import string
from datetime import datetime
from pathlib import Path
from shutil import copytree, copy2
from PyQt5.QtGui import QPixmap
import pandas as pd

from .logprint import print
from .rtsconfig import RTSconfig


class Report:

    def __init__(self, parent, ui, settings):
        self.TAB_COVERAGE = 0
        self.TAB_SITEINFOS = 1
        self.TAB_DAILY = 2
        self.TAB_HOURLY = 3
        self.TAB_10MIN = 4
        self.TAB_RMOB = 5
        self.TAB_CHRONO = 6
        self.TAB_RTS = 7
        self.TAB_TOTAL = 8

        self.DIV_SIDEBAR = 0
        self.DIV_DAILY = 1
        self.DIV_HOURLY = 2
        self.DIV_10MIN = 3
        self.DIV_RMOB = 4
        self.DIV_CHRONO = 5
        self.DIV_RTS = 6
        self.DIV_INDEX = 7
        self.DIV_TOTAL = 8

        self._sections = list(range(self.TAB_SITEINFOS, self.TAB_TOTAL))
        self._ui = ui
        self._parent = parent
        self._settings = settings
        self._dataSource = None
        self._classFilter = None
        self._RMOBfile = None
        self._itemSelected = None
        self._defaultPreface = None

        self._dirPath = None
        self._assetsPath = None
        self._imgPath = None
        self._pageHeader = None
        self._pageFooter = None
        self._pageHeaderRmob = None
        self._totalDays = 0
        self._totUnfiltered = 0
        self._stationLogo = None
        self._events = list()
        self._statImgs = list()

        self._rdf = [None] * self.TAB_TOTAL  # report dataframes (-->XSLX)
        self._div = [None] * self.DIV_TOTAL  # report divs (-->HTML)

        self._ui.pbRefresh_3.clicked.connect(self.updateTabReport)

        self._ui.chkOverdense_3.clicked.connect(self._setClassFilter)
        self._ui.chkUnderdense_3.clicked.connect(self._setClassFilter)
        self._ui.chkFakeEsd_3.clicked.connect(self._setClassFilter)
        self._ui.chkFakeCar1_3.clicked.connect(self._setClassFilter)
        self._ui.chkFakeCar2_3.clicked.connect(self._setClassFilter)
        self._ui.chkFakeSat_3.clicked.connect(self._setClassFilter)
        self._ui.chkFakeLong_3.clicked.connect(self._setClassFilter)
        self._ui.chkAll_3.clicked.connect(self._toggleCheckAll)

    def _recallDefaultPrefaceText(self):
        preface = self._defaultPreface
        self._ui.pteReportPreface.setPlainText(preface)

    def _recallDefaultStoredText(self):
        preface = self._settings.readSettingAsString("preface")
        if preface != "":
            preface = preface.replace("<br>", "\n")
        self._ui.pteReportPreface.setPlainText(preface)

    def updateTabReport(self):
        self._parent.busy(True)
        self._ui.pbReportXLSX.clicked.connect(self._reportXLSX)
        self._ui.pbReportHTML.clicked.connect(self._reportHTML)
        self._ui.pbRecallDefault.clicked.connect(self._recallDefaultPrefaceText)
        self._ui.pbRecallStored.clicked.connect(self._recallDefaultStoredText)
        self._ui.chkSiteInfosExc.setChecked(self._settings.readSettingAsBool('siteInfosExc'))
        self._ui.chkDailyExc.setChecked(self._settings.readSettingAsBool('dailyExc'))
        self._ui.chkRMOBexc.setChecked(self._settings.readSettingAsBool('RMOBexc'))
        self._ui.chkHourlyExc.setChecked(self._settings.readSettingAsBool('hourlyExc'))
        self._ui.chk10minExc.setChecked(self._settings.readSettingAsBool('tenMinExc'))
        self._ui.chkChronoExc.setChecked(self._settings.readSettingAsBool('chronoExc'))
        self._ui.chkExpEvExc.setChecked(self._settings.readSettingAsBool('expEvExc'))
        self._ui.chkExpStTabExc.setChecked(self._settings.readSettingAsBool('stTabExc'))
        self._ui.chkExpStGrpExc.setChecked(self._settings.readSettingAsBool('stGrpExc'))
        self._ui.chkSetupExc.setChecked(self._settings.readSettingAsBool('setupExc'))

        self._ui.chkSiteInfosExc.clicked.connect(self._toggleSection)
        self._ui.chkDailyExc.clicked.connect(self._toggleSection)
        self._ui.chkRMOBexc.clicked.connect(self._toggleSection)
        self._ui.chkHourlyExc.clicked.connect(self._toggleSection)
        self._ui.chk10minExc.clicked.connect(self._toggleSection)
        self._ui.chkChronoExc.clicked.connect(self._toggleSection)
        self._ui.chkExpEvExc.clicked.connect(self._toggleSection)
        self._ui.chkExpStTabExc.clicked.connect(self._toggleSection)
        self._ui.chkExpStGrpExc.clicked.connect(self._toggleSection)
        self._ui.chkSetupExc.clicked.connect(self._toggleSection)

        self._dataSource = self._parent.dataSource
        self._parent.updateStatusBar("Gathering informations for report...")
        self._getClassFilter()

        # creates a default preface

        self._defaultPreface = '''
        Radio meteor reception report produced by station {0}, located in {1}, {2}
        on days between {3} and {4}, generated with the <i>Echoes Data Browser</i> program.
        
        A total of {5} events have been received.
        
        The tables that can be recalled through the side menu <i>Tables</i> include
        the chronological of the events with the related details and counts, 
        which are represented with different time resolutions or daily by classification.
        The counts reported regard only the events satisfying the class filters set when
        this report has been generated. These filters are reported above each table.
        
        In addition to the tables mentioned above - which are automatically generated by the program - this
        report includes a set of screens, graphs and tables that have been exported manually 
        by the operator. This set is divided into two galleries, depending if they
        concern individuals events or statistical counts.
        
        Regarding the latter, keep in mind that exports can have been done in any previous
        moment before the report generation, so they could have been generated with different filters applied.
        For this reason, their comments always include the related filter settings.
        
        '''.format(self._settings.readSettingAsString("stationName"),
                   self._settings.readSettingAsString("country"),
                   self._settings.readSettingAsString("city"),
                   self._settings.readSettingAsString("dateFrom"),
                   self._settings.readSettingAsString("dateTo"),
                   self._dataSource.totalsUnclassified(self._settings.readSettingAsString("dateFrom"),
                                                       self._settings.readSettingAsString("dateTo"))
                   )

        preface = self._settings.readSettingAsString("preface")

        if preface != "":
            preface = preface.replace("<br>", "\n")
        else:
            preface = self._defaultPreface

        self._ui.pteReportPreface.setPlainText(preface)
        if self._parent.batchReport:
            # when self generated, ignores the personalized preface from ebrow.ini
            # and uses always a self generated one
            self._ui.pteReportPreface.setPlainText(self._defaultPreface)
            self._reportHTML()

        self._parent.busy(False)

    def _prepareTables(self):
        self._parent.busy(True)
        if self._rdf[self.TAB_COVERAGE] is None:
            self._buildCoverageTable()

        if self._rdf[self.TAB_SITEINFOS] is None:
            self._buildSiteInfosTable()

        if self._rdf[self.TAB_DAILY] is None:
            self._buildDailyCountsTable()

        if self._rdf[self.TAB_RMOB] is None:
            self._buildRMOBtable()

        if self._rdf[self.TAB_HOURLY] is None:
            self._buildHourlyCountsTable()

        if self._rdf[self.TAB_10MIN] is None:
            self._buildTenMinCountsTable()

        if self._rdf[self.TAB_CHRONO] is None:
            self._buildChronologicalTable()

        if self._rdf[self.TAB_RTS] is None:
            self._buildSetupTable()

        self._parent.busy(False)

    def _toggleSection(self):
        # self._sections.append(self.TAB_COVERAGE) # always included
        for (key, chk) in [('siteInfosExc', self._ui.chkSiteInfosExc),
                           ('RMOBexc', self._ui.chkRMOBexc),
                           ('dailyExc', self._ui.chkDailyExc),
                           ('hourlyExc', self._ui.chkHourlyExc),
                           ('tenMinExc', self._ui.chk10minExc),
                           ('chronoExc', self._ui.chkChronoExc),
                           ('expEvExc', self._ui.chkExpEvExc),
                           ('stTabExc', self._ui.chkExpStTabExc),
                           ('stGrpExc', self._ui.chkExpStGrpExc),
                           ('setupExc', self._ui.chkSetupExc)]:
            if not chk.isChecked():
                self._settings.writeSetting(key, False)
            else:
                self._settings.writeSetting(key, True)

    def _getClassFilter(self):
        self._parent.busy(True)
        self._classFilter = self._settings.readSettingAsString('classFilterReport')
        idx = 0
        for tag in self._parent.filterTags:
            isCheckTrue = tag in self._classFilter
            if idx < len(self._parent.filterCheckReport):
                self._parent.filterCheckReport[idx].setChecked(isCheckTrue)
                idx += 1
        self._parent.updateStatusBar("Filtering report data by {}".format(self._classFilter))
        self._parent.busy(False)

    def _setClassFilter(self):
        self._parent.busy(True)
        classFilter = ''
        idx = 0
        for check in self._parent.filterCheckReport:
            if check.isChecked():
                classFilter += self._parent.filterTags[idx] + ','
            idx += 1

        if classFilter != '':
            classFilter = classFilter[0:-1]  # discards latest comma+space

        self._classFilter = classFilter
        self._settings.writeSetting('classFilterReport', self._classFilter)
        self._parent.updateStatusBar("Filtering report data by classification: {}".format(self._classFilter))
        self._parent.busy(False)

    def _toggleCheckAll(self):
        self._ui.chkOverdense_3.setChecked(self._ui.chkAll_3.isChecked())
        self._ui.chkUnderdense_3.setChecked(self._ui.chkAll_3.isChecked())
        self._ui.chkFakeEsd_3.setChecked(self._ui.chkAll_3.isChecked())
        self._ui.chkFakeCar1_3.setChecked(self._ui.chkAll_3.isChecked())
        self._ui.chkFakeCar2_3.setChecked(self._ui.chkAll_3.isChecked())
        self._ui.chkFakeSat_3.setChecked(self._ui.chkAll_3.isChecked())
        self._ui.chkFakeLong_3.setChecked(self._ui.chkAll_3.isChecked())
        self._setClassFilter()

    def _fillImportList(self):

        # scanning exported event-related files
        shotDir = Path(self._parent.exportDir, 'events')
        fileList = os.listdir(shotDir)

        # sort the filenames by date then by ID
        toSort = list()
        for fileName in fileList:
            fields = fileName.split('_')
            patchedFileName = "{}_{}${}".format(fields[-2], fields[-1], fileName)
            toSort.append(patchedFileName)
        toSort.sort()
        fileList = []
        for patchedFileName in toSort:
            fields = patchedFileName.split('$')
            fileList.append(fields[1])

        # create a list of event-related files (screenshots/graphs/comments)
        self._events = []
        detFile = None
        powFile = None
        img2dFile = None
        img3dFile = None
        commFile = None

        for fileName in fileList:
            img = Path(fileName)
            # looking for optional auxiliary files
            detFile = None
            powFile = None
            img2dFile = None
            img3dFile = None
            commFile = None

            if img.name.startswith('autoshot'):
                details = fileName.replace('autoshot', 'details')
                details = details.replace('png', 'csv')
                detPath = shotDir / Path(details)
                if detPath.exists():
                    detFile = detPath.name

                power = fileName.replace('autoshot', 'power_profile')
                powPath = shotDir / Path(power)
                if powPath.exists():
                    powFile = powPath.name

                img2d = fileName.replace('autoshot', 'image2d')
                img2dPath = shotDir / Path(img2d)
                if img2dPath.exists():
                    img2dFile = img2dPath.name

                img3d = fileName.replace('autoshot', 'image3d')
                img3dPath = shotDir / Path(img3d)
                if img3dPath.exists():
                    img3dFile = img3dPath.name

                comments = 'comments_' + fileName
                comments = comments.replace('png', 'txt')
                commPath = shotDir / Path(comments)
                if commPath.exists():
                    commFile = commPath.name

                imgFile = fileName
                htmlFile = imgFile.replace('.png', '.html')
                evTuple = (imgFile, htmlFile, commFile, detFile, powFile, img2dFile, img3dFile)
                self._events.append(evTuple)

        # scanning exported statistic files (tables/graphs/comments)
        statDir = Path(self._parent.exportDir, 'statistics')
        fileList = os.listdir(statDir)
        fileList.sort()

        # create a list of event-related files
        self._statImgs = []

        for fileName in fileList:
            dataFile = Path(fileName)
            # statistic graph
            if dataFile.suffix == '.png' and fileName.startswith('stat-'):
                comments = fileName.replace('stat-', 'comments-')
                comments = comments.replace('png', 'txt')
                commPath = statDir / Path(comments)
                commFile = None
                if commPath.exists():
                    commFile = commPath.name

                htmlFile = fileName.replace('.png', '.html')
                self._statImgs.append((dataFile, htmlFile, commFile))

        self._statTabs = []
        for fileName in fileList:
            dataFile = Path(fileName)
            # statistic table
            if dataFile.suffix == '.csv' and fileName.startswith('stat-'):
                comments = fileName.replace('stat-', 'comments-')
                comments = comments.replace('csv', 'txt')
                commPath = statDir / Path(comments)
                commFile = None
                if commPath.exists():
                    commFile = commPath.name

                htmlFile = fileName.replace('.csv', '.html')
                self._statTabs.append((dataFile, htmlFile, commFile))

    def _buildCoverageTable(self):
        self._parent.updateStatusBar("Gathering time coverage data...")

        # first row contains current UTC date and time
        utc = datetime.utcnow()
        (fromStr, toStr) = self._settings.coverage()

        # note: this is not a real uptime
        df = self._dataSource.getASpartialFrame(self._parent.fromDate, self._parent.toDate)
        deltaMins = df['delta_min'].sum()
        coveredIDs = self._parent.covID

        # main header
        rdfCols = ['A', 'B', 'C', 'D', 'E', 'F', 'G']
        rdfDict = dict()
        rdfDict['1'] = [':/rts', 'Meteor scatter detection report', ':/logo']

        # section 1 report coverage, cannot be excluded
        rdfDict['2'] = ['Report coverage']
        rdfDict['3'] = ['Created by', 'Echoes Data Browser', 'version:', self._parent.version, ' on ',
                        utc.strftime("%Y %m %d"), utc.strftime("%H:%M:%S")]
        rdfDict['4'] = ['Covered period from: ', fromStr, ' to: ', toStr, 'total events:', coveredIDs]
        rdfDict['5'] = ['Complessive uptime: ', deltaMins, ' minutes']

        rdf = pd.DataFrame.from_dict(rdfDict, orient='index', columns=rdfCols)
        self._rdf[self.TAB_COVERAGE] = rdf

    def _buildSiteInfosTable(self):
        self._parent.updateStatusBar("Gathering site informations...")
        if not self._ui.chkSiteInfosExc.isChecked():
            rdfCols = ['A', 'B', 'C', 'D', 'E', 'F']
            rdfDict = dict()
            for rowNr in range(1, len(rdfCols)):
                rdfDict[str(rowNr)] = len(rdfCols) * [' ']
            # main header

            # section 2 site infos
            self._stationLogo = self._settings.readSettingAsString('logoPath')
            if self._stationLogo == ':/logo':
                self._stationLogo = 'logo.png'
            rdfDict['1'] = ['Site infos']
            rdfDict['2'] = ['Station name: ', self._settings.readSettingAsString('stationName'),
                            self._stationLogo]
            rdfDict['3'] = ['Owner data: ', self._settings.readSettingAsString('owner')]
            rdfDict['4'] = ['Country: ', self._settings.readSettingAsString('country'), 'City: ',
                            self._settings.readSettingAsString('city')]
            rdfDict['5'] = ['Longitude: ', self._settings.readSettingAsString('longitude'), 'Latitude: ',
                            self._settings.readSettingAsString('latitude')]
            rdfDict['6'] = ['Altitude: ', self._settings.readSettingAsString('altitude')]
            rdfDict['7'] = ['Antenna type: ', self._settings.readSettingAsString('antenna'),
                            'Azimuth: ', str(self._settings.readSettingAsString('antAzimuth')) + '\xb0',
                            'Elevation: ', str(self._settings.readSettingAsString('antElevation')) + '\xb0']

            rdfDict['8'] = ['Preamplifier: ', self._settings.readSettingAsString('preamplifier')]
            rdfDict['9'] = ['Receiver: ', self._settings.readSettingAsString('receiver')]
            rdfDict['10'] = ['Frequencies: ', self._settings.readSettingAsString('frequencies')]
            rdfDict['11'] = ['Computer type: ', self._settings.readSettingAsString('computer')]
            rdfDict['12'] = ['e-mail: ', self._settings.readSettingAsString('email')]
            rdfDict['13'] = ['Notes: ', self._settings.readSettingAsString('notes')]

            rdf = pd.DataFrame.from_dict(rdfDict, orient='index', columns=rdfCols)
        else:
            rdf = None
        self._rdf[self.TAB_SITEINFOS] = rdf

    def _buildRMOBtable(self):

        if not self._ui.chkRMOBexc.isChecked():
            self._parent.updateStatusBar("Gathering RMOB countings...")
            # section 3 hourly counts by month (RMOB data)
            dfMonth, heatmapFileName, bargraphFileName = self._parent.getRMOBdata(sendOk=False)

            # in order to merge the daily counts table with the page header, their column names
            # must become a data row and column names become standardized numbers

            colNames = dfMonth.columns.values.tolist()  # gets the column names
            dfMonth.columns = range(dfMonth.columns.size)  # reset column names

            # creates a dataframe containing the column names as single row
            dfCols = pd.DataFrame([colNames], columns=list(range(0, len(colNames))))

            # the dailyDf index contains the dates, they must become a data column
            dates = dfMonth.index.values.tolist()
            dfMonth.insert(0, 'Day', dates)
            dfMonth.columns = range(dfMonth.columns.size)  # reset column names

            # here creates the page header dataframe
            columns = len(dfMonth.columns)

            # section 4 hourly counts by month, RMOB format
            # now joins the three dataframes
            fullDf = pd.concat([dfCols, dfMonth], ignore_index=True)
            fullDf.columns = list(string.ascii_uppercase + string.ascii_lowercase)[
                             :fullDf.columns.size]  # replace column names with letter
            fullDf.loc[0, 'B':] = fullDf.loc[0, 'A':].shift(periods=1, fill_value='')
            fullDf.loc[0, 'A'] = 'DAY'
            title = pd.DataFrame([['Hourly counts by month, RMOB format', '', '', '', '', '', '', '', '', '', '', '',
                                   '', '', '', '', '', '', '', '', '', '', '', '', '']], columns=fullDf.columns)
            fullDf = pd.concat([title, fullDf], ignore_index=True)

            # RMOBimgName = self._settings.readSettingAsString("RMOBfilePrefix") + ".jpg"
            # RMOBfullPath = Path(self._parent.exportDir, "statistics", RMOBimgName)

            # fullDf.loc[2, 'B'] = str(RMOBfullPath.absolute())

            # nex index must start from 1, not zero
            fullDf.index = fullDf.index + 1
        else:
            fullDf = None

        self._rdf[self.TAB_RMOB] = fullDf

    def _buildDailyCountsTable(self):
        if not self._ui.chkDailyExc.isChecked():
            self._parent.updateStatusBar("Gathering daily counts...")

            df = self._dataSource.getADpartialFrame(self._parent.fromDate, self._parent.toDate)

            dailyDf = self._dataSource.dailyByClassification(df, self._classFilter,
                                                             self._parent.fromDate, self._parent.toDate,
                                                             totalRow=True, totalColumn=True)

            # in order to merge the daily counts table with the page header, their column names
            # must become a data row and column names become standardized numbers

            colNames = ['Day'] + dailyDf.columns.values.tolist()  # gets the column names
            dailyDf.columns = range(dailyDf.columns.size)  # reset column names

            # creates a dataframe containing the column names as single row
            dfCols = pd.DataFrame([colNames], columns=list(range(0, len(colNames))))

            # the dailyDf index contains the dates, they must become a data column
            dates = dailyDf.index.values.tolist()
            dailyDf.insert(0, 'Day', dates)
            dailyDf.reset_index(inplace=True, drop=True)
            dailyDf.columns = range(dailyDf.columns.size)  # reset column names

            # here creates the page header dataframe
            columns = len(dailyDf.columns)
            rdfCols = list(range(0, columns))
            rdfDict = dict()
            rdfDict['1'] = len(rdfCols) * [' ']

            # section 3 daily counts by classifications

            rdf = pd.DataFrame.from_dict(rdfDict, orient='index', columns=rdfCols)

            # now joins the three dataframes
            fullDf = pd.concat([rdf, dfCols, dailyDf], ignore_index=True)

            fullDf.columns = list(string.ascii_uppercase +
                                  string.ascii_lowercase)[:fullDf.columns.size]  # replace column names with letters

            # the column name 'DAY' is prepended for correct columns/headers alignment
            fullDf.loc[0, 'A'] = 'Daily counts by classification'
            # nex index must start from 1, not zero
            fullDf.index = fullDf.index + 1
        else:
            fullDf = None

        self._rdf[self.TAB_DAILY] = fullDf

    def _buildHourlyCountsTable(self):
        if not self._ui.chkHourlyExc.isChecked():
            self._parent.updateStatusBar("Gathering hourly counts...")
            df = self._dataSource.getADpartialFrame(self._parent.fromDate, self._parent.toDate)

            hourlyDf = self._dataSource.makeDaily(df, self._parent.fromDate, self._parent.toDate, dtRes='H',
                                                  filters=self._classFilter)

            # in order to merge the daily counts table with the page header, their column names
            # must become a data row and column names become standardized numbers
            colNames = hourlyDf.columns.values.tolist()  # gets the column names
            hourlyDf.columns = range(hourlyDf.columns.size)  # reset column names

            # creates a dataframe containing the column names as single row
            dfCols = pd.DataFrame([colNames], columns=list(range(0, len(colNames))))

            # the dailyDf index contains the dates, they must become a data column
            dates = hourlyDf.index.values.tolist()
            hourlyDf.insert(0, 'Day', dates)
            hourlyDf.columns = range(hourlyDf.columns.size)  # reset column names

            # here creates the page header dataframe
            columns = len(hourlyDf.columns)
            rdfCols = list(range(0, columns))
            rdfDict = dict()
            rdfDict['1'] = len(rdfCols) * [' ']

            # section 3 Hourly counts
            rdf = pd.DataFrame.from_dict(rdfDict, orient='index', columns=rdfCols)

            # now joins the three dataframes
            fullDf = pd.concat([rdf, dfCols, hourlyDf], ignore_index=True)

            fullDf.columns = list(string.ascii_uppercase)[:fullDf.columns.size]  # replace column names with letters
            # the column name 'DAY' is prepended for correct columns/headers alignment
            fullDf.loc[0, 'A'] = 'Hourly counts'
            fullDf.loc[1, 'B':] = fullDf.loc[1, 'A':].shift(periods=1, fill_value=' ')
            fullDf.loc[1, 'A'] = 'utc_date'
            # nex index must start from 1, not zero
            fullDf.index = fullDf.index + 1
        else:
            fullDf = None
        self._rdf[self.TAB_HOURLY] = fullDf

    def _buildTenMinCountsTable(self):
        if not self._ui.chk10minExc.isChecked():
            self._parent.updateStatusBar("Gathering 10min counts...")
            df = self._dataSource.getADpartialFrame(self._parent.fromDate, self._parent.toDate)

            tenDf = self._dataSource.makeDaily(df, self._parent.fromDate, self._parent.toDate, dtRes='10T',
                                               filters=self._classFilter)
            tenDf = self._dataSource.splitAndStackDataframe(tenDf, maxColumns=24)

            # in order to merge the daily counts table with the page header, their column names
            # must become a data row and column names become standardized numbers
            colNames = tenDf.columns.values.tolist()  # gets the column names
            tenDf.columns = range(tenDf.columns.size)  # reset column names

            # creates a dataframe containing the column names as single row
            # dfCols = pd.DataFrame([colNames], columns=list(range(0, len(colNames))))

            # the dailyDf index contains the dates, they must become a data column
            # dates = tenDf.index.values.tolist()
            # tenDf.insert(0, 'Day', dates)
            # tenDf.columns = range(tenDf.columns.size)  # reset column names

            # here creates the page header dataframe
            columns = len(tenDf.columns)
            rdfCols = list(range(0, columns))
            rdfDict = dict()
            rdfDict['1'] = len(rdfCols) * [' ']

            # section 3 Counts by 10 min intervals
            rdf = pd.DataFrame.from_dict(rdfDict, orient='index', columns=rdfCols)

            # now joins the three dataframes

            fullDf = pd.concat([rdf, tenDf], ignore_index=True)

            fullDf.columns = list(string.ascii_uppercase + string.ascii_lowercase)[
                             :fullDf.columns.size]  # replace column names with letters
            # the column name 'DAY' is prepended for correct columns/headers alignment
            # fullDf.loc[3, 'A':] = fullDf.loc[3, 'A':].shift(periods=1, fill_value='DAY')
            # nex index must start from 1, not zero
            # fullDf.index = fullDf.index + 1
            fullDf.reset_index(drop=True)
            fullDf.index = fullDf.index + 1
            rows = len(fullDf.index) + 1
            days = int(rows / 6)

            fullDf.iloc[0, 0] = 'Counts by 10 min intervals'

            # fullDf.drop(columns='INDEX', inplace=True)
        else:
            fullDf = None
        self._rdf[self.TAB_10MIN] = fullDf

    def _buildChronologicalTable(self):

        if not self._ui.chkChronoExc.isChecked():
            self._parent.updateStatusBar("Getting chronological events table...")
            df = self._parent.dataSource.getADpartialCompositeFrame(self._parent.fromDate, self._parent.toDate,
                                                                    self._classFilter)
            self._rdf[self.TAB_CHRONO] = df

    def _buildSetupTable(self):
        if not self._ui.chkSetupExc.isChecked():
            self._parent.updateStatusBar("Getting Echoes RTS setup table...")
            revList = self._dataSource.getCfgRevisions()
            # considers only the latest revision
            rev = revList[-1]
            rts = RTSconfig(self._parent, self._ui, self._settings)
            self._rdf[self.TAB_RTS] = rts.getAllTables(rev)

    def _reportXLSX(self):
        self._prepareTables()
        self._parent.busy(True)
        # create a excel writer object
        self._parent.updateStatusBar("Saving XLSX report file")
        with pd.ExcelWriter("report.xlsx") as writer:
            # use to_excel function and specify the sheet_name and index
            # to store the dataframe in specified sheet
            # fullRdf.to_excel(writer, sheet_name="Report", index=False)
            self._parent.updateStatusBar("Adding Time Coverage sheet to XLSX...")
            self._rdf[self.TAB_COVERAGE].to_excel(writer, sheet_name="Time coverage", index=False, header=None)

            if not self._ui.chkSiteInfosExc.isChecked():
                self._parent.updateStatusBar("Adding Site Informations sheet to XLSX...")
                self._rdf[self.TAB_SITEINFOS].to_excel(writer, sheet_name="Site information", index=False, header=None)

            if not self._ui.chkDailyExc.isChecked():
                self._parent.updateStatusBar("Adding Daily Counts sheet to XLSX...")
                self._rdf[self.TAB_DAILY].to_excel(writer, sheet_name="Daily counts", index=False, header=None)

            if not self._ui.chkHourlyExc.isChecked():
                self._parent.updateStatusBar("Adding Hourly Counts sheet to XLSX...")
                self._rdf[self.TAB_HOURLY].to_excel(writer, sheet_name="Hourly counts", index=False, header=None)

            if not self._ui.chk10minExc.isChecked():
                self._parent.updateStatusBar("Adding 10min Counts sheet to XLSX...")
                self._rdf[self.TAB_10MIN].to_excel(writer, sheet_name="10min counts", index=False, header=None)

            if not self._ui.chkRMOBexc.isChecked():
                self._parent.updateStatusBar("Adding Current RMOB sheet to XLSX...")
                self._rdf[self.TAB_RMOB].to_excel(writer, sheet_name="current month's RMOB", index=False)

            if not self._ui.chkChronoExc.isChecked():
                self._parent.updateStatusBar("Adding Chronological table sheet to XLSX...")
                self._rdf[self.TAB_CHRONO].to_excel(writer, sheet_name="chronological events table", index=False)

            if not self._ui.chkSetupExc.isChecked():
                self._parent.updateStatusBar("Adding RTS setup tables sheets to XLSX...")
                sheets = ['RTS device configuration',
                          'RTS FFT configuration',
                          'RTS Output configuration',
                          'RTS Storage configuration',
                          'RTS Preferences'
                          ]
                idx = 0
                for table in self._rdf[self.TAB_RTS]:
                    sheet = sheets[idx]
                    table.to_excel(writer, sheet_name=sheet, index=False)
                    idx += 1

        # TODO: openpyxl is able to embed images also, besides data
        # but I need to know a way to get the Workbook created by to_excel()
        # to continue work on it, or otherwise reopen the xslx as Workbook
        # and continue working on it, adding the images and comments.
        self._parent.busy(False)

    def _reportHTML(self):
        self._parent.busy(True)
        self._prepareTables()
        self._parent.updateStatusBar("Saving HTML report files")
        now = datetime.utcnow()
        dirName = now.strftime("report_%Y-%m-%d_%H-%M-%S")
        self._dirPath = Path(".") / Path(dirName)
        self._assetsPath = self._dirPath / "assets"

        # self._cssPath = assetsPath / "css"
        # self._jsPath = assetsPath / "js"
        # cssPath.mkdir(mode=0o644, parents=True, exist_ok=True)
        # imgPath.mkdir(mode=0o644, parents=True, exist_ok=True)
        # jsPath.mkdir(mode=0o644, parents=True, exist_ok=True)

        # copies the html assets directory tree with its files
        self._imgPath = self._assetsPath / "img"
        srcAssetsPath = Path(self._parent.appPath) / "assets" / "img"
        print("Copying from {} to {}".format(srcAssetsPath, self._assetsPath))
        copytree(srcAssetsPath, self._imgPath)
        self._cssPath = self._assetsPath / "css"
        srcAssetsPath = Path(self._parent.appPath) / "assets" / "css"
        print("Copying from {} to {}".format(srcAssetsPath, self._assetsPath))
        copytree(srcAssetsPath, self._cssPath)

        # the site logo is copied into report's assets img
        if self._stationLogo != 'logo.png':
            copy2(self._stationLogo, self._imgPath)
        # if the logo is the default one, there is no need to copy
        # since its file is already present under assets/

        df = self._rdf[self.TAB_DAILY]
        if df is None:
            self._parent.updateStatusBar("Cannot create HTML reports, all sections are excluded")
            self._parent.busy(False)
            return

        self._totalDays = len(df.index) - 3
        if self._parent.covID < 2 or self._totalDays < 2:
            message = "Cannot generate a report covering less than 2 days"
            self._parent.updateStatusBar(message)
            if not self._parent.batchReport:
                self._parent.infoMessage("Report generation", message)
            sys.exit()

        df = self._rdf[self.TAB_COVERAGE]
        self._totUnfiltered = df.loc['4', 'F']

        rmobCSSpatch = '<link rel="stylesheet" href="assets/css/rmob.css">'
        headerModel = '''<!DOCTYPE html>
<html lang="en">
    <head>
        <title>Echoes Report</title>
        <!-- <link rel="stylesheet" href="assets/css/reset.css"> -->
        <link rel="stylesheet" href="assets/css/style.css">
        <link rel="stylesheet" href="assets/css/table.css">
        {0}
        <style>
            @import url('https://fonts.googleapis.com/css2?family=Titillium+Web&display=swap');
        </style>
    </head>
    <body onload="onLoadFunc()">
        <div id="loader"></div>
        <div id="container">
            <div class="info clearfix">
                <img src="assets/img/logo.png" alt="Echoes">
                <h1>Meteor scatter detection report</h1>
                <p>Created with <i>Echoes Data Browser</i> release {1}</p>
            </div> <!-- info -->
            <div id="content" class="animate-bottom">
'''

        self._pageHeader = headerModel.format('', self._parent.version)
        self._pageHeaderRmob = headerModel.format(rmobCSSpatch, self._parent.version)

        self._pageFooter = '''\t\t\t<div id="footer">
                <div class="box">
                    <strong>Echoes</strong> Reporting System - <a href="https://www.gabb.it/echoes" title="Echoes Web Site">Web Site</a> -
                     <a href="https://sourceforge.net/projects/echoes/" title="Find Echoes on Sourceforge">Find Echoes on Sourceforge</a> - 
                     <a href="https://www.facebook.com/gmbertani/" title="Find Echoes on Facebook">Find Echoes on Facebook</a>
                </div><!--box-->
            </div> <!-- footer -->
        </div> <!-- container -->
        <script>
            var tout;
            
            function onLoadFunc() {
                tout = setTimeout(showPage, 3000);
            }
            
            function showPage() {
                document.getElementById("loader").style.display = "none";
                document.getElementById("content").style.display = "flex";
            }
        </script>
    </body>
</html>'''

        self._fillImportList()
        self._parent.updateStatusBar("Embedding exported event files")
        for evTup in self._events:
            self._buildEventHtml(evTup)

        self._parent.updateStatusBar("Embedding exported statistic graphs")
        for stTup in self._statImgs:
            self._buildStatGraphsHtml(stTup)
        self._parent.updateStatusBar("Embedding exported statistic tables")
        for tabTup in self._statTabs:
            self._buildStatTablesHtml(tabTup)

        self._buildIndexHtml('index.html')
        self._buildDailyHtml('daily.html')
        self._buildHourlyHtml('hourly.html')
        self._build10minHtml('10min.html')
        self._buildRMOBhtml('rmob.html')
        self._buildChronologicalHtml('chrono.html')
        self._buildRTShtml('setup.html')
        self._parent.busy(False)

    def _buildIndexHtml(self, outputFile):
        self._parent.updateStatusBar("Creating index.html")
        summary = self._parent.getSummaryPlot(self._classFilter)
        srcImgFilePath = Path(self._parent.workingDir) / summary
        copy2(srcImgFilePath, self._imgPath)
        relImgPath = self._imgPath.relative_to(self._dirPath)
        summaryPath = relImgPath / summary
        summaryPath = summaryPath.as_posix()

        preface = self._ui.pteReportPreface.toPlainText()
        preface = preface.replace('\n', '<br>')

        self._div[self.DIV_INDEX] = '''\t\t\t<div class="box">
                <h2>Preface</h2>
                <p>{0}</p>
                <hr>
                <h2>Daily Summary plot</h2>
'''.format(preface)
        self._div[self.DIV_INDEX] += self._activeFiltersHtml()
        self._div[self.DIV_INDEX] += '''<img src = "{0}" title = "Daily summary plot">
            </div><!--box-->
        </div><!--content-->
'''.format(summaryPath)

        self._settings.writeSetting("preface", preface)
        self._buildSidebarHtml(outputFile)

        htmlDoc = self._pageHeader
        htmlDoc += self._div[self.DIV_INDEX]
        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / outputFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _buildHourlyHtml(self, outputFile):
        if self._ui.chkHourlyExc.isChecked():
            return

        self._parent.updateStatusBar("Creating hourly.html")
        df = self._rdf[self.TAB_HOURLY]

        totals = df.iloc[-1, -1]

        self._div[self.DIV_HOURLY] = '''\t\t\t<div class="box">
                    <h2 class="graph">Daily statistic counts</h2>
                      <p>Total filtered events displayed: <strong>{0}</strong> of <strong>{1}</strong> unfiltered.</p>
                    <h3>1 hour resolution</h3>            
'''.format(totals, self._totUnfiltered)

        self._div[self.DIV_HOURLY] += self._activeFiltersHtml()
        self._div[self.DIV_HOURLY] += '''\t\t\t\t<table class="blueTable">
                        <thead>
                            <tr>
'''
        for col in range(0, len(df.columns)):
            self._div[self.DIV_HOURLY] += '\t\t\t\t\t\t\t<th>{}</th>\n'.format(df.iloc[1, col])
        self._div[self.DIV_HOURLY] += '''\t\t\t\t\t\t</tr>
                        </thead>
                        <tbody>
'''
        for row in range(2, len(df.index)):
            self._div[self.DIV_HOURLY] += '\t\t\t\t\t\t<tr>\n'
            for col in range(0, len(df.columns)):
                self._div[self.DIV_HOURLY] += '\t\t\t\t\t\t\t<td>{}</td>\n'.format(df.iloc[row, col])
            self._div[self.DIV_HOURLY] += '\t\t\t\t\t\t</tr>\n'
        self._div[self.DIV_HOURLY] += '''\t\t\t\t\t</tbody>
                    </table>
                </div><!--box-->
            </div><!--content-->
'''
        self._buildSidebarHtml(outputFile)

        htmlDoc = self._pageHeader
        htmlDoc += self._div[self.DIV_HOURLY]
        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / outputFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _buildChronologicalHtml(self, outputFile):
        """
        Being a very long and wide table, it must be stored in a frame
        to allow browsing via scrollbars
        """
        if self._ui.chkChronoExc.isChecked():
            return

        self._parent.updateStatusBar("Creating chrono.html")
        iframeHeader = '''<!DOCTYPE html>
<html lang="en">
    <head>
        <title>Chronological event\'s table</title>
        <!-- <link rel="stylesheet" href="assets/css/reset.css"> -->
        <link rel="stylesheet" href="assets/css/style.css">
        <link rel="stylesheet" href="assets/css/table.css">
        <style>
            @import url('https://fonts.googleapis.com/css2?family=Titillium+Web&display=swap');
        </style>
    </head>
    <body>
        <div id="container">
            <div class="info clearfix">
                <h2>Chronological event\'s table</h2>
            </div> <!-- info -->
            <div id="frametable">
'''

        df = self._rdf[self.TAB_CHRONO]
        frameHtml = iframeHeader
        frameHtml += df.to_html(index=False)
        frameHtml += self._pageFooter

        tableIFrame = "table-" + outputFile
        outFilePath = self._dirPath / tableIFrame
        with open(outFilePath, 'w') as f:
            f.write(frameHtml)

        self._buildSidebarHtml(outputFile)

        self._div[self.DIV_CHRONO] = '''\t\t\t\t<iframe src="{}" title="chrono table frame" height="2000"></iframe>
            </div><!--content-->\n
'''.format(tableIFrame)

        htmlDoc = self._pageHeader
        htmlDoc += self._div[self.DIV_CHRONO]
        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / outputFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _buildDailyHtml(self, outputFile):
        if self._ui.chkDailyExc.isChecked():
            return

        self._parent.updateStatusBar("Creating daily.html")
        df = self._rdf[self.TAB_DAILY]
        totals = df.iloc[-1, -1]

        self._div[self.DIV_DAILY] = '''\t\t\t<div class="box">
                <h2 class="graph">Daily statistic counts by classifications</h2>
                <p>Total filtered events displayed: <strong>{0}</strong> of <strong>{1}</strong> unfiltered.</p>
                <h3>1 day resolution</h3>            
'''.format(totals, self._totUnfiltered)
        self._div[self.DIV_DAILY] += self._activeFiltersHtml()
        self._div[self.DIV_DAILY] += '''\t\t\t\t<table class="blueTable">
                    <thead>
                        <tr>
                            <th>UTC date</th>
'''
        if self._isChecked('OVER') == 'checked':
            self._div[self.DIV_DAILY] += '\t\t\t\t\t\t\t<th>Overdense</th>\n'
        if self._isChecked('UNDER') == 'checked':
            self._div[self.DIV_DAILY] += '\t\t\t\t\t\t\t<th>Underdense</th>\n'
        if self._isChecked('FAKE ESD') == 'checked':
            self._div[self.DIV_DAILY] += '\t\t\t\t\t\t\t<th>Fake ESD</th>\n'
        if self._isChecked('FAKE CAR1') == 'checked':
            self._div[self.DIV_DAILY] += '\t\t\t\t\t\t\t<th>Fake Carrier 1</th>\n'
        if self._isChecked('FAKE CAR2') == 'checked':
            self._div[self.DIV_DAILY] += '\t\t\t\t\t\t\t<th>FaKe Carrier 2</th>\n'
        if self._isChecked('FAKE SAT') == 'checked':
            self._div[self.DIV_DAILY] += '\t\t\t\t\t\t\t<th>Fake Saturation</th>\n'
        if self._isChecked('FAKE LONG') == 'checked':
            self._div[self.DIV_DAILY] += '\t\t\t\t\t\t\t<th>Fake Long</th>\n'
        self._div[self.DIV_DAILY] += '\t\t\t\t\t\t\t<th>Daily total</th>\n'
        self._div[self.DIV_DAILY] += '''\t\t\t\t\t\t</tr>
                    </thead>
                    <tbody>
'''
        for row in range(2, len(df.index)):
            self._div[self.DIV_DAILY] += '\t\t\t\t\t\t<tr>\n'
            for col in range(0, len(df.columns)):
                self._div[self.DIV_DAILY] += '\t\t\t\t\t\t\t<td>{}</td>\n'.format(df.iloc[row, col])
            self._div[self.DIV_DAILY] += '\t\t\t\t\t\t</tr>\n'

        self._div[self.DIV_DAILY] += '''\t\t\t\t\t</tbody>
                </table>
            </div><!--box-->
        </div><!--content-->
'''
        self._buildSidebarHtml(outputFile)

        htmlDoc = self._pageHeader
        htmlDoc += self._div[self.DIV_DAILY]
        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / outputFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _buildHourlyHtml(self, outputFile):
        if self._ui.chkHourlyExc.isChecked():
            return
        df = self._rdf[self.TAB_HOURLY]

        totals = df.iloc[-1, -1]

        self._div[self.DIV_HOURLY] = '''\t\t\t<div class="box">
                    <h2 class="graph">Daily statistic counts</h2>
                      <p>Total filtered events displayed: <strong>{0}</strong> of <strong>{1}</strong> unfiltered.</p>
                    <h3>1 hour resolution</h3>            
'''.format(totals, self._totUnfiltered)

        self._div[self.DIV_HOURLY] += self._activeFiltersHtml()
        self._div[self.DIV_HOURLY] += '''\t\t\t\t<table class="blueTable">
                        <thead>
                            <tr>
'''
        for col in range(0, len(df.columns)):
            self._div[self.DIV_HOURLY] += '\t\t\t\t\t\t\t<th>{}</th>\n'.format(df.iloc[1, col])
        self._div[self.DIV_HOURLY] += '''\t\t\t\t\t\t</tr>
                        </thead>
                        <tbody>
'''
        for row in range(2, len(df.index)):
            self._div[self.DIV_HOURLY] += '\t\t\t\t\t\t<tr>\n'
            for col in range(0, len(df.columns)):
                self._div[self.DIV_HOURLY] += '\t\t\t\t\t\t\t<td>{}</td>\n'.format(df.iloc[row, col])
            self._div[self.DIV_HOURLY] += '\t\t\t\t\t\t</tr>\n'
        self._div[self.DIV_HOURLY] += '''\t\t\t\t\t</tbody>
                    </table>
                </div><!--box-->
            </div><!--content-->
'''
        self._buildSidebarHtml(outputFile)

        htmlDoc = self._pageHeader
        htmlDoc += self._div[self.DIV_HOURLY]
        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / outputFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _build10minHtml(self, outputFile):
        if self._ui.chk10minExc.isChecked():
            return

        self._parent.updateStatusBar("Creating 10min.html")
        df = self._rdf[self.TAB_10MIN]
        totals = df.iloc[-1, -1]

        self._div[self.DIV_10MIN] = '''\t\t\t<div class="box">
                <h2 class="graph">Daily statistic counts</h2>
                  <p>Total filtered events displayed: <strong>{0}</strong> of <strong>{1}</strong> unfiltered.</p>
                <h3>10 min resolution</h3>            
'''.format(totals, self._totUnfiltered)

        self._div[self.DIV_10MIN] += self._activeFiltersHtml()
        self._div[self.DIV_10MIN] += '\t\t\t\t<table class="blueTable">\n'

        for row in range(1, len(df.index)):
            if row % (self._totalDays + 2) == 1:
                # heading row
                self._div[self.DIV_10MIN] += '''\t\t\t\t\t<thead>
                        <tr>
'''
                for col in range(0, len(df.columns)):
                    self._div[self.DIV_10MIN] += '\t\t\t\t\t\t\t<th>{}</th>\n'.format(df.iloc[row, col])
                self._div[self.DIV_10MIN] += '''\t\t\t\t\t\t</tr>
                    </thead>
'''
            else:
                if row % (self._totalDays + 2) == 2:
                    # begin data body
                    self._div[self.DIV_10MIN] += '\t\t\t\t\t<tbody>\n'

                # data row
                self._div[self.DIV_10MIN] += '\t\t\t\t\t\t<tr>\n'
                for col in range(0, len(df.columns)):
                    self._div[self.DIV_10MIN] += '\t\t\t\t\t\t\t<td>{}</td>\n'.format(df.iloc[row, col])
                self._div[self.DIV_10MIN] += '\t\t\t\t\t\t</tr>\n'

                if row % (self._totalDays + 2) == self._totalDays:
                    # end data body
                    self._div[self.DIV_10MIN] += '\t\t\t\t\t<tbody>\n'

        self._div[self.DIV_10MIN] += '''\t\t\t\t</table>
            </div><!--box-->
        </div><!--content-->
'''
        self._buildSidebarHtml(outputFile)

        htmlDoc = self._pageHeader
        htmlDoc += self._div[self.DIV_10MIN]
        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / outputFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _buildRMOBhtml(self, outputFile):
        if self._ui.chkRMOBexc.isChecked():
            return

        self._parent.updateStatusBar("Creating rmob.html")
        df = self._rdf[self.TAB_RMOB]
        self._RMOBfile = self._settings.readSettingAsString('RMOBfilePrefix') + ".jpg"
        srcImgFilePath = Path(self._parent.workingDir) / self._RMOBfile
        copy2(srcImgFilePath, self._imgPath)
        relImgPath = self._imgPath.relative_to(self._dirPath)
        RMOBimgPath = relImgPath / self._RMOBfile
        RMOBimgPath = RMOBimgPath.as_posix()

        self._div[self.DIV_RMOB] = '''\t\t\t<div class="box">
                <h2>RMOB summary picture</h2>
                    <img src = "{0}" title = "RMOB summary">
                <h2 class="graph">Current month's Daily statistic counts - RMOB format</h2>           
'''.format(RMOBimgPath)
        self._div[self.DIV_RMOB] += '''\t\t\t\t<table class="blueTable">
                        <thead>
                            <tr>
'''
        for col in range(0, len(df.columns)):
            self._div[self.DIV_RMOB] += '\t\t\t\t\t\t\t<th>{}</th>\n'.format(df.iloc[1, col])
        self._div[self.DIV_RMOB] += '''\t\t\t\t\t\t</tr>
                        </thead>
                        <tbody>
'''
        for row in range(2, len(df.index)):
            self._div[self.DIV_RMOB] += '\t\t\t\t\t\t<tr>\n'
            for col in range(0, len(df.columns)):
                self._div[self.DIV_RMOB] += '\t\t\t\t\t\t\t<td>{}</td>\n'.format(df.iloc[row, col])
            self._div[self.DIV_RMOB] += '\t\t\t\t\t\t</tr>\n'
        self._div[self.DIV_RMOB] += '''\t\t\t\t\t</tbody>
                    </table>

            </div><!--box-->
        </div><!--content-->
'''

        self._buildSidebarHtml(outputFile)

        htmlDoc = self._pageHeaderRmob
        htmlDoc += self._div[self.DIV_RMOB]
        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / outputFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _buildRTShtml(self, outputFile):
        if self._ui.chkSetupExc.isChecked():
            return

        self._parent.updateStatusBar("Creating setup.html")
        dfList = self._rdf[self.TAB_RTS]

        cfgFile = self._dataSource.stem()
        if 'snapshot_' in cfgFile:
            cfgFile = cfgFile.replace('snapshot_', '')

        self._div[self.DIV_RTS] = '''\t\t\t<div class="box">
                <h2 class="graph">Echoes settings</h2>
                <p>Settings stored in configuration file <strong>{0}.rts</strong></p>
'''.format(cfgFile)

        titles = ['\t\t\t\t\t<h3>Device settings</h3>\n',
                  '\t\t\t\t\t<h3>FFT settings</h3>\n',
                  '\t\t\t\t\t<h3>Output settings</h3>\n',
                  '\t\t\t\t\t<h3>Storage settings</h3>\n',
                  '\t\t\t\t\t<h3>Preferences</h3>\n']

        idx = 0

        for df in dfList:
            self._div[self.DIV_RTS] += titles[idx]
            self._div[self.DIV_RTS] += '''\t\t\t\t\t<table class="blueTable">
                        <thead>
                            <tr>
'''
            # header
            for col in df.columns:
                self._div[self.DIV_RTS] += '\t\t\t\t\t\t\t<th>{}</th>\n'.format(col)
            self._div[self.DIV_RTS] += '''\t\t\t\t\t\t</tr>
                            </thead>
                            <tbody>
'''
            # data
            for row in range(0, len(df.index)):
                self._div[self.DIV_RTS] += '\t\t\t\t\t\t<tr>\n'
                for col in range(0, len(df.columns)):
                    self._div[self.DIV_RTS] += '\t\t\t\t\t\t\t<td>{}</td>\n'.format(df.iloc[row, col])
                self._div[self.DIV_RTS] += '\t\t\t\t\t\t</tr>\n'
            self._div[self.DIV_RTS] += '''\t\t\t\t\t</tbody>
                    </table>
                    <hr>
'''
            idx += 1

        self._div[self.DIV_RTS] += '''\t\t\t</div><!--box-->
        </div><!--content-->        
'''

        self._buildSidebarHtml(outputFile)

        htmlDoc = self._pageHeader
        htmlDoc += self._div[self.DIV_RTS]
        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / outputFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _buildEventHtml(self, evTuple):
        if self._ui.chkExpEvExc.isChecked():
            return
        (imgFile, htmlFile, commFile, detFile, powFile, img2dFile, img3dFile) = evTuple
        srcFilePath = Path(self._parent.exportDir) / "events"

        srcImgFilePath = srcFilePath / imgFile
        copy2(srcImgFilePath, self._imgPath)

        srcPowFilePath = None
        if powFile is not None:
            srcPowFilePath = srcFilePath / powFile
            copy2(srcPowFilePath, self._imgPath)

        srcImg2dFilePath = None
        if img2dFile is not None:
            srcImg2dFilePath = srcFilePath / img2dFile
            copy2(srcImg2dFilePath, self._imgPath)

        srcImg3dFilePath = None
        if img3dFile is not None:
            srcImg3dFilePath = srcFilePath / img3dFile
            copy2(srcImg3dFilePath, self._imgPath)

        relImgPath = self._imgPath.relative_to(self._dirPath)
        imgFilePath = relImgPath / imgFile

        # comments and details files remain under the export directory
        # to be integrated into the page as HTML, while images are copied
        # under page's assets directory, so the report directory can
        # be moved or compressed without caring of external resources

        # reads comments if present
        comment = "\n"
        if commFile is not None:
            commFilePath = srcFilePath / commFile
            with open(commFilePath, 'r') as f:
                for line in f.readlines():
                    line = '\t\t\t\t\t' + line.strip() + '<br>\n'
                    comment += line

        # reads the event details table if present
        df = None
        if detFile is not None:
            detFilePath = srcFilePath / detFile
            # reads the csv in a dataframe
            df = pd.read_csv(detFilePath, sep=self._settings.dataSeparator())

        self._buildSidebarHtml(imgFile)

        dt, id = imgFile[-20:-4].split('_')
        htmlDoc = self._pageHeader
        htmlDoc += '''\t\t\t\t<div class="box">
                    <h2 class="graph">Events gallery, event#{0}</h2>
                    <p>{1}</p>
                    <hr>
                    <h3>Screenshot</h3>
                        <img src="{2}" alt="{3}">
                    <hr>
'''.format(id, comment, imgFilePath.as_posix(), imgFilePath.name)

        if powFile is not None:
            powFilePath = relImgPath / powFile
            htmlDoc += '''\t\t\t\t\t<h3>Total power profile</h3>
                        <img src="{0}" alt="{1}">
                    <hr>
'''.format(powFilePath.as_posix(), powFilePath.name)

        if img2dFile is not None:
            img2dFilePath = relImgPath / img2dFile
            htmlDoc += '''\t\t\t\t\t<h3>2D data plotting</h3>
                        <img src="{0}" alt="{1}">
                    <hr>
'''.format(img2dFilePath.as_posix(), img2dFilePath.name)

        if img3dFile is not None:
            img3dFilePath = relImgPath / img3dFile
            htmlDoc += '''\t\t\t\t\t<h3>3D data plotting</h3>
                        <img src="{0}" alt="{1}">
                    <hr>
'''.format(img3dFilePath.as_posix(), img3dFilePath.name)

        if df is not None:
            htmlDoc += '''\t\t\t\t\t<h3>Event details</h3>
                        <table class="blueTable" style="table-layout: fixed;word-wrap: break-word;">
                            <thead>
                                <tr>
'''
            for col in df.columns:
                if "Unnamed: 0" in col:
                    col = " "
                htmlDoc += '\t\t\t\t\t\t\t\t<th>{}</th>\n'.format(col)
            htmlDoc += '''\t\t\t\t\t\t\t\t\t</tr>
                        </thead>
                        <tbody>
'''
            for row in range(0, len(df.index)):
                htmlDoc += '\t\t\t\t\t\t\t<tr>\n'
                for col in range(0, len(df.columns)):
                    if col == 0:
                        htmlDoc += '\t\t\t\t\t\t\t\t<td><strong>{}</strong></td>\n'.format(df.iloc[row, 0])
                    else:
                        htmlDoc += '\t\t\t\t\t\t\t\t<td>{}</td>\n'.format(df.iloc[row, col])
                htmlDoc += '\t\t\t\t\t\t\t</tr>\n'
            htmlDoc += '''\t\t\t\t\t\t</tbody>
                    </table>
                </div><!--box-->
            </div><!--content-->
'''
        else:
            htmlDoc += '''\t\t\t\t</div><!--box-->
            </div><!--content-->
'''
        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / htmlFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _buildStatTablesHtml(self, stTup):
        if self._ui.chkExpStTabExc.isChecked():
            return

        # reads the comments file if existing
        comment = "\n"
        if stTup[2] is not None:
            txtFile = str(stTup[2])
            txtFilePath = Path(self._parent.exportDir) / "statistics" / txtFile
            with open(txtFilePath, 'r') as f:
                for line in f.readlines():
                    line = '\t\t\t\t\t' + line.strip() + '<br>\n'
                    comment += line
            comment += '\t\t\t\t\t'

        # reads the csv in a dataframe
        tableFile = str(stTup[0])
        outputFile = str(stTup[1])
        srcTabFilePath = Path(self._parent.exportDir) / "statistics" / tableFile
        df = pd.read_csv(srcTabFilePath, sep=self._settings.dataSeparator())

        self._buildSidebarHtml(outputFile)

        # builds the page
        htmlDoc = self._pageHeader
        htmlDoc += '''\t\t\t\t<div class="box">
                            <h2 class="data">Statistic data gallery</h2>
                            <p>{0}</p>
'''.format(comment)

        htmlDoc += '''\t\t\t\t<table class="blueTable">
                                <thead>
                                    <tr>
'''
        for col in df.columns:
            if "Unnamed: 0" in col:
                col = " "
            htmlDoc += '\t\t\t\t\t\t\t<th>{}</th>\n'.format(col)
        htmlDoc += '''\t\t\t\t\t\t</tr>
                                </thead>
                            <tbody>
'''
        for row in range(0, len(df.index)):
            htmlDoc += '\t\t\t\t\t\t<tr>\n'
            for col in range(0, len(df.columns)):
                htmlDoc += '\t\t\t\t\t\t\t<td>{}</td>\n'.format(df.iloc[row, col])
            htmlDoc += '\t\t\t\t\t\t</tr>\n'
        htmlDoc += '''\t\t\t\t\t</tbody>
                            </table>
                        </div><!--box-->
                    </div><!--content-->
        '''

        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / outputFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _buildStatGraphsHtml(self, stTup):
        if self._ui.chkExpStGrpExc.isChecked():
            return
        imageFile = str(stTup[0])
        outputFile = str(stTup[1])
        srcImgFilePath = Path(self._parent.exportDir) / "statistics" / imageFile
        copy2(srcImgFilePath, self._imgPath)
        relImgPath = self._imgPath.relative_to(self._dirPath)
        imgFilePath = relImgPath / imageFile
        comment = "\n"
        if stTup[2] is not None:
            txtFile = str(stTup[2])
            txtFilePath = Path(self._parent.exportDir) / "statistics" / txtFile
            with open(txtFilePath, 'r') as f:
                for line in f.readlines():
                    line = '\t\t\t\t\t' + line.strip() + '<br>\n'
                    comment += line
            comment += '\t\t\t\t\t'
        self._buildSidebarHtml(imageFile)
        htmlDoc = self._pageHeader
        htmlDoc += '''\t\t\t\t<div class="box">
                    <h2 class="graph">Statistic data gallery</h2>
                    <p>{0}</p>
                    <img src="{1}" alt="{2}">
                </div>
            </div><!--content-->
'''.format(comment, imgFilePath.as_posix(), imageFile)

        htmlDoc += self._div[self.DIV_SIDEBAR]
        htmlDoc += self._pageFooter

        outFilePath = self._dirPath / outputFile
        with open(outFilePath, 'w') as f:
            f.write(htmlDoc)

    def _buildSidebarHtml(self, currentFile):
        self._div[self.DIV_SIDEBAR] = '\t\t\t<div id="sidebar">\n'

        df = self._rdf[self.TAB_COVERAGE]
        if df is None:
            self._parent.updateStatusBar("skipping excluded section")
            return

        if currentFile != 'index.html':
            divBoxBack = '''\t\t\t\t\t<div class="box index">
                        <a href="index.html" title="Back to index" class="current"><strong>Back to Index</strong></a>
                    </div><!-- box -->
'''
        else:
            divBoxBack = ''

        divBoxDate = '''\t\t\t\t<div class="box date">
                    <h2 class="calendar">Time span</h2>
                    <dl>
                        <dt>From</dt>        
                        <dd>{0}</dd>
                        <dt>To</dt>
                        <dd>{1}</dd>
                        <dt>Total</dt>
                        <dd>{2} days</dd>
                        <dt>Uptime</dt>
                        <dd>{3} min.</dd>
                    </dl>
                </div><!--date-->
'''.format(df.loc['4', 'B'], df.loc['4', 'D'], self._totalDays, df.loc['5', 'B'])

        df = self._rdf[self.TAB_SITEINFOS]

        origLogoPath = df.loc['2', 'C']
        origLogo = QPixmap(origLogoPath)
        destPath = self._imgPath / origLogoPath
        origLogoName = destPath.name
        origLogo.save(origLogoName)
        logoPath = Path("assets/img") / origLogoName

        divSiteInfo = '''\t\t\t\t<div class="box">
                    <h2 class="site">Site information</h2>
                    <img src="{0}" alt="Site logo not found" width="128">  
                    <dl>
                        <dt>Station name</dt>
                        <dd>{1}</dd>
                        <dt>Owner</dt>
                        <dd>{2}</dd>
                        <dt>Country</dt>
                        <dd>{3}</dd>
                        <dt>City</dt>
                        <dd>{4}</dd>
                        <dt>Longitude</dt>
                        <dd>{5}</dd>
                        <dt>Latitude</dt>
                        <dd>{6}</dd>
                        <dt>Altitude</dt>
                        <dd>{7}m</dd>
                    </dl>
                </div><!-- box -->
'''.format(
            logoPath.as_posix(),  # logo
            df.loc['2', 'B'],  # station name
            df.loc['3', 'B'],  # owner
            df.loc['4', 'B'], df.loc['4', 'D'],  # country, city
            df.loc['5', 'B'], df.loc['5', 'D'],  # longitude, latitude
            df.loc['6', 'B']  # altitude
        )

        divReceiver = '''\t\t\t\t<div class="box">
                    <h2 class="radio">Receiver details</h2>
                    <dl>
                        <dt>Antenna type</dt>
                        <dd>{0}</dd>
                        <dt>Azimuth</dt>
                        <dd>{1}</dd>
                        <dt>Elevation</dt>
                        <dd>{2}</dd>
                        <dt>Preamplifier</dt>
                        <dd>{3}</dd>
                        <dt>Receiver</dt>
                        <dd>{4}</dd>
                        <dt>Frequencies</dt>
                        <dd>{5} Hz</dd>
                        <dt>Computer type</dt>
                        <dd>{6}</dd>
                        <dt>Software</dt>
                        <dd>Echoes v.{7} + Ebrow v.{8}</dd>
                        <dt>e-mail</dt>
                        <dd>{9}</dd>
                        <dt>Notes:</dt>
                        <dd>{10}</dd>
                    </dl>
                </div><!-- box -->
'''.format(df.loc['7', 'B'], df.loc['7', 'D'], df.loc['7', 'F'],
           df.loc['8', 'B'], df.loc['9', 'B'], df.loc['10', 'B'],
           df.loc['11', 'B'], '0.54++', self._parent.version,
           df.loc['12', 'B'], df.loc['13', 'B'])  # TODO: the echoes version must be embedded in DB

        divTables = '''\t\t\t\t<div class="box">
                    <h2 class="data">Tables</h2>
                    <ul>
'''
        if not self._ui.chkDailyExc.isChecked():
            divTables += '\t\t\t\t\t\t<li><a href="daily.html" title="Daily counts by classification" {}>Daily ' \
                         'counts by classifications</a></li>\n' \
                .format('class="current"' if currentFile == 'daily.html' else '')
        if not self._ui.chkHourlyExc.isChecked():
            divTables += '\t\t\t\t\t\t<li><a href="hourly.html" title="Hourly counts by day" {}>Hourly counts by ' \
                         'day</a></li>\n' \
                .format('class="current"' if currentFile == 'hourly.html' else '')
        if not self._ui.chk10minExc.isChecked():
            divTables += '\t\t\t\t\t\t<li><a href="10min.html" title="10 minutes counts by day" {}>10 minutes counts ' \
                         'by day</a></li>\n '.format('class="current"' if currentFile == '10min.html' else '')
        if not self._ui.chkRMOBexc.isChecked():
            divTables += '\t\t\t\t\t\t<li><a href="rmob.html"  title="Current month\'s RMOB counts" {}>Current ' \
                         'month\'s RMOB counts</a></li>\n '.format(
                'class="current"' if currentFile == 'rmob.html' else '')
        if not self._ui.chkChronoExc.isChecked():
            divTables += '\t\t\t\t\t\t<li><a href="chrono.html"  title="Event\'s chronological table" {}>' \
                         'Event\'s chronological table</a></li>\n '.format(
                'class="current"' if currentFile == 'chrono.html' else '')
        if not self._ui.chkSetupExc.isChecked():
            divTables += '\t\t\t\t\t\t<li><a href="setup.html"  title="Echoes setup tables" {}>' \
                         'Echoes setup tables (RTS)</a></li>\n '.format(
                'class="current"' if currentFile == 'setup.html' else '')
        divTables += '''
                    </ul>
                </div><!-- box -->
'''

        divEvents = '''\t\t\t\t<div class="box">
                    <h2 class="events">Exported events images gallery</h2>
                    <ul>
'''

        if self._ui.chkExpEvExc.isChecked():
            divEvents += '\t\t\t\t\t\t<li>None</li>\n'
        else:
            for evTuple in self._events:
                (imgFile, htmlFile, commFile, detFile, powFile, img2dFile, img3dFile) = evTuple
                dt, id = imgFile[-20:-4].split('_')
                isCurrent = 'class="current"' if currentFile == imgFile else ''
                divEvents += '\t\t\t\t\t\t<li><a href="{0}" title="{1}" {4}>Screenshot {2}, {3}</a></li>\n'.format(
                    htmlFile, imgFile, id, dt, isCurrent)

        divEvents += '''
                    </ul>
                </div><!-- box -->
'''

        divStats = '''\t\t\t\t<div class="box">
                    <h2 class="graph">Exported statistics data gallery</h2>
                    <ul>
'''
        if self._ui.chkExpStTabExc.isChecked():
            divStats += '\t\t\t\t\t\t<li>None</li>\n'
        else:
            for stTup in self._statImgs:
                imgName = str(stTup[0])
                htmlName = str(stTup[1])
                fields = imgName.split('-')
                kind = fields[-1].split('.')[0]
                title = fields[1].replace('_', ' ')
                resolution = fields[2].replace('_', ' ')
                prog = fields[3]
                isCurrent = 'class="current"' if currentFile == imgName else ''
                divStats += '\t\t\t\t\t\t<li><a href="{0}" title="{1}" {6}>{2} {3} {4} #{5}</a></li>\n'.format(
                    htmlName, imgName, kind, title, resolution, prog, isCurrent)

            for tabTup in self._statTabs:
                tabName = str(tabTup[0])
                htmlName = str(tabTup[1])
                fields = tabName.split('-')
                kind = fields[-1].split('.')[0]
                title = fields[1].replace('_', ' ')
                resolution = fields[2].replace('_', ' ')
                prog = fields[3]
                isCurrent = 'class="current"' if currentFile == tabName else ''
                divStats += '\t\t\t\t\t\t<li><a href="{0}" title="{1}" {6}>{2} {3} {4} #{5}</a></li>\n'.format(
                    htmlName, tabName, kind, title, resolution, prog, isCurrent)

        divStats += '''
                    </ul>
                </div><!-- box -->
'''

        self._div[self.DIV_SIDEBAR] += divBoxBack + '\n'
        self._div[self.DIV_SIDEBAR] += divBoxDate + '\n'
        self._div[self.DIV_SIDEBAR] += divTables + '\n'
        self._div[self.DIV_SIDEBAR] += divEvents + '\n'
        self._div[self.DIV_SIDEBAR] += divStats + '\n'
        self._div[self.DIV_SIDEBAR] += divSiteInfo + '\n'
        self._div[self.DIV_SIDEBAR] += divReceiver + '\n'
        self._div[self.DIV_SIDEBAR] += '\t\t\t</div><!--sidebar-->\n'

    def _isChecked(self, filt):
        if filt in self._classFilter:
            return 'checked'
        return ''

    def _filterDescription(self, filt):
        descDict = dict()
        descDict['OVER'] = 'Overdense'
        descDict['UNDER'] = 'Underdense'
        descDict['FAKE ESD'] = 'Fake ESD'
        descDict['FAKE CAR1'] = 'Fake Carrier 1'
        descDict['FAKE CAR2'] = 'Fake Carrier 2'
        descDict['FAKE SAT'] = 'Fake Saturation'
        descDict['FAKE LONG'] = 'Fake Long'
        return descDict[filt]

    def _activeFiltersHtml(self):
        html = '''\t\t\t\t<h4>Active filters:</h4>
                <ol>
                    <li> 
                        <input type="checkbox" id="od" name="od" value="Overdense events" {0} onclick="return false;">
                        <label for="od">Overdense events</label>
                    </li>
                    <li> 
                        <input type="checkbox" id="ud" name="ud" value="Underdense events" {1} onclick="return false;">
                        <label for="ud">Underdense events</label>
                    </li>
                    <li> 
                        <input type="checkbox" id="esd" name="esd" value="Fake events (ESD)" {2} onclick="return false;">
                        <label for="esd">Fake events (ESD)</label>
                    </li>
                    <li> 
                        <input type="checkbox" id="car1" name="car1" value="Fake events (Carrier type 1)" {3} onclick="return false;">
                        <label for="car1">Fake events (Carrier type 1)</label>
                    </li>
                    <li> 
                        <input type="checkbox" id="car2" name="car2" value="Fake events (Carrier type 2)" {4} onclick="return false;">
                        <label for="car2">Fake events (Carrier type 2)</label>
                    </li>
                    <li> 
                        <input type="checkbox" id="sat" name="sat" value="Fake events (RX saturation)" {5} onclick="return false;">
                        <label for="car2">Fake events (RX saturation)</label>
                    </li>
                    <li> 
                        <input type="checkbox" id="long" name="long" value="Fake events (too long lasting)" {6} onclick="return false;">
                        <label for="car2">Fake events (too long lasting)</label>
                    </li>
                </ol>
'''.format(self._isChecked('OVER'), self._isChecked('UNDER'), self._isChecked('FAKE ESD'),
           self._isChecked('FAKE CAR1'),
           self._isChecked('FAKE CAR2'), self._isChecked('FAKE SAT'), self._isChecked('FAKE LONG'))
        return html



"""
******************************************************************************

    Echoes Data Browser (Ebrow) is a data navigation and report generation
    tool for Echoes.
    Echoes is a RF spectrograph for SDR devices designed for meteor scatter
    Both copyright (C) 2018-2023
    Giuseppe Massimo Bertani gm_bertani(a)yahoo.it

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html

*******************************************************************************

"""
import os
import shutil
import numpy as np
from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtCore import QByteArray

import pandas as pd
import math
import matplotlib.ticker as ticker
import matplotlib as mp
from matplotlib.colors import ListedColormap
from ctypes import *
from matplotlib.dates import num2date
from pathlib import Path
from .logprint import print


class DumpHeader(Structure):
    _pack_ = 1
    _fields_ = [('startMarker', 4 * c_char),
                ('resolution', c_double),
                ('startHz', c_uint32),
                ('points', c_uint32),
                ('isAudio', c_bool)]


class ScanHeader(Structure):
    _pack_ = 1
    _fields_ = [('scanMarker', 2 * c_char),
                ('timestamp_us', c_double),
                ('firstCell', c_float)]


class ScanFooter(Structure):
    _pack_ = 1
    _fields_ = [('lastCell', c_float),
                ('endMarker', 2 * c_char),
                ('avgN', c_float),
                ('maxDbfs', c_float),
                ('maxDiff', c_float)]


class PrecisionDateFormatter(ticker.Formatter):
    """
    Extend the `matplotlib.ticker.Formatter` class to allow for millisecond
    precision when formatting a tick (in days since the epoch) with a
    `~datetime.datetime.strftime` format string.

    """

    def __init__(self, fmt, precision=3, tz=None):
        """
        Parameters
        ----------
        fmt : str
            `~datetime.datetime.strftime` format string.
        """

        if tz is None:
            from matplotlib.dates import _get_rc_timezone
            tz = _get_rc_timezone()

        self.fmt = fmt
        self.tz = tz
        self.precision = precision

    def __call__(self, x, pos=0):
        # if x == 0:
        #     raise ValueError("PrecisionDateFormatter found a value of 0, which is "
        #                      "an illegal date; this usually occurs because "
        #                      "you have not informed that axis that it is "
        #                      "plotting dates, e.g., with ax.xaxis_date()")

        dt = num2date(x, self.tz)
        ms = dt.strftime("%f")[:self.precision]

        return dt.strftime(self.fmt).format(ms=ms)

    def set_tzinfo(self, tz):
        self.tz = tz


def splitASCIIdumpFile(data: QByteArray):
    lineNum = 1
    dataStr = data.data().decode()
    dataLines = dataStr.splitlines()
    mapList = []
    powerList = []
    for line in dataLines:
        cells = line.split()
        if len(cells) >= 3:
            timeMs = float(cells[0])  # int(float(cells[0]) * 1000)
            if len(cells) == 3:
                cells[0] = timeMs
                cells[1] = int(cells[1])
                cells[2] = float(cells[2])
                mapList.append(cells[:3])

            if len(cells) == 7:
                # scan line end
                cells[0] = timeMs
                cells[1] = int(cells[1])
                cells[2] = float(cells[2])
                cells[3] = timeMs
                cells[4] = float(cells[4])
                cells[5] = float(cells[5])
                cells[6] = float(cells[6])
                powerList.append(cells[3:])

    dfMap = pd.DataFrame(mapList, columns=['time', 'frequency', 'S'])
    dfPower = pd.DataFrame(powerList, columns=['time', 'N', 'S', 'diff'])
    # dfPower['time'] = (pd.to_datetime(dfPower['time'], unit='s'))
    dfPower['timestamp'] = (pd.to_datetime(dfPower['time'], unit='s'))
    # dfMap['time'] = dfMap['time'].astype('datetime64')

    return dfMap, dfPower


def splitBinaryDumpFile(rawData: QByteArray):
    """

    """
    dh = DumpHeader()
    print("sizeof dh=", sizeof(dh))
    sh = ScanHeader()
    print("sizeof sh=", sizeof(sh))
    sf = ScanFooter()
    print("sizeof sf=", sizeof(sf))

    beg = 0
    end = beg + sizeof(dh)

    dh = DumpHeader.from_buffer(rawData[beg:end])
    if dh.startMarker != b'DUMP':
        print("binary dump file out of sync: ", dh.startMarker)
        return None, None

    scanNr = 0
    mapList = []
    powerList = []
    crc_16 = c_uint16

    while end < (rawData.length() - sizeof(crc_16)):
        sh = ScanHeader()
        beg = end
        end += sizeof(sh)
        sh = ScanHeader.from_buffer(rawData[beg:end])
        if sh.scanMarker != b'$$':
            print("binary dump file, scan#{} out of sync: {}".format(scanNr, sh.scanMarker))
            return None, None

        # print("reading scan#", scanNr)

        # the first cell value is in the header
        mapList.append([sh.timestamp_us, dh.startHz, sh.firstCell])

        # while other cells must be read as array of [dh.points - 2] elements
        # since the first cell has been read with the header and
        # the last one will be read with the footer

        class ScanArray(Array):
            _length_ = (dh.points - 2)
            _type_ = c_float

        beg = end
        end += sizeof(ScanArray)
        scanArr = ScanArray.from_buffer(rawData[beg:end])

        for idx in range(0, dh.points - 2):
            point = idx + 1
            freq = dh.startHz + (dh.resolution * point)
            dBfs = scanArr[idx]
            mapList.append([sh.timestamp_us, freq, dBfs])

        beg = end
        end += sizeof(sf)
        sf = ScanFooter.from_buffer(rawData[beg:end])
        if sf.endMarker != b'&&':
            print("binary dump file, scan#{} termination out of sync: {}".format(scanNr, sf.endMarker))
            return None, None

        powerList.append([sh.timestamp_us, sf.avgN, sf.maxDbfs, sf.maxDiff])
        scanNr += 1

    # TODO: CRC-16 check
    if end < rawData.length():
        beg = end
        end += sizeof(crc_16)
        crc_16 = c_uint16.from_buffer(rawData[beg:end]).value
        print("CRC-16 read: {:04x}".format(crc_16))
    else:
        print("*** MISSING CRC16 ***")

    dfMap = pd.DataFrame(mapList, columns=['time', 'frequency', 'S'])
    dfPower = pd.DataFrame(powerList, columns=['time', 'N', 'S', 'diff'])
    dfPower['timestamp'] = (pd.to_datetime(dfPower['time'], unit='s'))
    print("binary data read successful")
    return dfMap, dfPower


def castFloatPrecision(value: float, precision: int = 4):
    if value != 0:
        multipliers = [0, 10, 100, 1000, 10e3, 10e4, 10e5, 10e6, 10e7]
        if precision > 8:
            precision = 8
        mult = multipliers[precision]
        v2 = value * mult
        result = int(v2) / mult
        # print ("cast: {} --> {}".format(value, result))
    else:
        result = value
    return result


def clearLayout(layout):
    toDelete = layout.count()
    print("Clearing {} thumbnails".format(toDelete))
    while layout.count() > 0:
        child = layout.takeAt(0)
        if child.widget():
            child.widget().hide()
            child.widget().deleteLater()
            print("deleted child widget")
        elif child.layout():
            clearLayout(child.layout())
            print("deleted child layout")


def notice(title: str, msg: str):
    errorbox = QMessageBox()
    errorbox.setWindowTitle("Echoes data browser")
    errorbox.setText(title + "\n" + msg)
    errorbox.setIcon(QMessageBox.Information)
    errorbox.exec_()


def getColorMap(name: str = 'echoes'):
    # echoes color scale
    echoes = [
        "black",
        "blue",
        "purple",
        "orange",
        "yellow",
        "white"
    ]

    # Colorgramme-like 24 colors scale
    # the black means -1 == data missing that is not
    # included in colormap but set at display time
    colorgramme = [
        "#0064fe",
        "#006dfd",
        "#0079f0",
        "#0087fe",
        "#009eff",
        "#00b0ff",
        "#00c6f3",
        "#00eaff",
        "#38fcf5",
        "#4efdd4",
        "#6cfcb8",
        "#83fca1",
        "#a4f892",
        "#b8fc7a",
        "#d1fc68",
        "#f4f96d",
        "#ffe600",
        "#ffc500",
        "#ffae00",
        "#ff9400",
        "#ff8600",
        "#ff7300",
        "#ff6300",
        "#ff1600"
    ]

    cmap = None
    if name == 'echoes':
        colors = echoes
        totColors = len(colors)
    elif name == 'colorgramme':
        colors = colorgramme
        totColors = len(colors)
    else:
        cmap = mp.colormaps[name]
        totColors = cmap.N
        # cmap = mp.colormaps[name]._resample(totColors)

    if cmap is None:
        cmap = ListedColormap(colors, name=name)
    print("name={}, total colors={}".format(name, totColors))
    return cmap


def createCustomColormapsDict():
    myCmaps = dict()

    # matplotlib embedded colormaps
    for key in mp.colormaps.keys():
        # print("key=", key)
        myCmaps[key] = mp.colormaps[key]

    # customized colormaps
    for key in ['echoes', 'colorgramme']:
        cmap = getColorMap(key)
        myCmaps[key] = cmap
    return myCmaps


def fpCmp(val1: float, val2: float, decimals: int):
    multiplier = math.pow(10, decimals)
    i1 = int(val1 * multiplier)
    i2 = int(val2 * multiplier)
    return i2 - i1


def fuzzyCompare(a: float, b: float, tolerance: float):
    diff = math.fabs(a - b)
    percDiff = math.fabs((diff * 100) / a)
    if percDiff > tolerance:
        return 1 if (a > b) else -1
    return 0


def normalize(a: object, beg: float, end: float, default: float):
    b = np.interp(a, (a.min(), a.max()), (beg, end))
    b[np.isnan(b)] = default
    return b


def cleanFolder(folderPath: str, removeEmptyFolders: bool = True):
    for filename in os.listdir(folderPath):
        file_path = os.path.join(folderPath, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path) and removeEmptyFolders:
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))
            return False
    return True


def mkExportFolder(folderPath: Path):
    return Path.mkdir(folderPath, parents=True, exist_ok=True)


def cryptDecrypt(st: str, key: int) -> str:
    stCrypt = ''
    for i in range(len(st)):
        stCrypt += chr(ord(st[i]) ^ (key >> 8))
    return stCrypt

def timestamp2sideral(ts: str) -> str:
    jd = pd.Timestamp(ts).to_julian_date()

    # sideral time in degrees
    gmst0 = (280.46061837 + 360.98564736629 * (jd - 2451545)) % 360

    # conversion to decimal hours:
    stDecHour = gmst0 / 15.0

    # extracting hexagesimal minutes:
    stHexaMin = math.modf(stDecHour)[0] * 60

    # extracting hexagesimal seconds:
    stHexaSec = math.modf(stHexaMin)[0] * 60

    # formatting string
    sideralTime = "{:02}:{:02}:{:02}".format(int(stDecHour), int(stHexaMin), int(stHexaSec))
    return sideralTime





"""
******************************************************************************

    Echoes Data Browser (Ebrow) is a data navigation and report generation
    tool for Echoes.
    Echoes is a RF spectrograph for SDR devices designed for meteor scatter
    Both copyright (C) 2018-2023
    Giuseppe Massimo Bertani gm_bertani(a)yahoo.it

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html

*******************************************************************************

"""

import pandas as pd
import numpy as np
import matplotlib as mp
import matplotlib.pyplot as plt
import matplotlib.dates as md
from mplcursors import cursor
from matplotlib.ticker import MultipleLocator
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from scipy.signal import savgol_filter
from .settings import Settings
from .basegraph import BaseGraph


mp.use('Qt5Agg')


class StatPlot(BaseGraph):

    def __init__(self, series: pd.Series, settings: Settings, inchWidth: float, inchHeight: float, title: str,
                 res: str = 'hour', showValues: bool = False, showGrid: bool = False,  smooth: bool = True,
                 subtractBackground: bool = False ):
        BaseGraph.__init__(self, settings)

        self._series = series
        x = self._series.index
        y = self._series.values

        windowSize = 17

        if smooth and y.size > windowSize:
            # smoothing by Savitsky-Golay filter
            # is switched off when data aren't enough

            if windowSize > y.size:
                windowSize = y.size

            print("windowsize=", windowSize)

            polyOrder = 5
            if polyOrder >= y.size:
                polyOrder = y.size-1
            print("polyOrder=", polyOrder)
            title += "\n Smoothed through Savitsky-Golay filter, windowSize={}, polyOrder={}".format(windowSize, polyOrder)
            y = savgol_filter(y, windowSize, polyOrder)

        colors = self._settings.readSettingAsObject('colorDict')
        backColor = colors['background'].name()
        plt.figure(figsize=(inchWidth, inchHeight))
        self._fig, ax = plt.subplots(1, facecolor=backColor)
        ax.set_facecolor(backColor)
        xdt = np.asarray(x, dtype='datetime64[s]')
        ax.xaxis.set_major_locator(MultipleLocator(1))
        myFmt = md.DateFormatter('%Y-%m-%d')
        ax.xaxis.set_major_formatter(myFmt)

        if res == '10m':
            if subtractBackground:
                title += "\nafter sporadic background subtraction"

            ax.plot(xdt, y, linestyle=self._settings.readSettingAsString('dataLineStyle'),
                    linewidth=self._settings.readSettingAsString('dataLineWidth'), color=colors['counts'].name(),
                    label='Counts')

            locator = MultipleLocator(6 / 144)
            locator.MAXTICKS = 10000
            ax.xaxis.set_minor_locator(locator)
            myFmt = md.DateFormatter('%Hh')
            ax.xaxis.set_minor_formatter(myFmt)
            ax.set_title(title, loc='left')

        elif res == 'hour':
            if subtractBackground:
                title += "\nafter sporadic background subtraction"

            ax.plot(xdt, y, linestyle=self._settings.readSettingAsString('dataLineStyle'),
                    linewidth=self._settings.readSettingAsString('dataLineWidth'), color=colors['counts'].name(),
                    label='Counts')

            ax.xaxis.set_minor_locator(MultipleLocator(6 / 24))
            myFmt = md.DateFormatter('%Hh')
            ax.xaxis.set_minor_formatter(myFmt)
            ax.set_title(title, loc='left')

        elif res == 'day':
            if subtractBackground:
                title += "\nafter sporadic background subtraction"

            ax.plot(xdt, y, linestyle=self._settings.readSettingAsString('dataLineStyle'),
                    linewidth=self._settings.readSettingAsString('dataLineWidth'), color=colors['counts'].name(),
                    label='Counts')

            ax.set_title(title, loc='left')

        ax.tick_params(axis='x', which='both', labelrotation=60)

        if showGrid:
            ax.grid(color=colors['majorGrids'].name(), linestyle=self._settings.readSettingAsString('majorLineStyle'),
                    linewidth=self._settings.readSettingAsString('majorLineWidth'), which='major')
            ax.grid(color=colors['minorGrids'].name(), linestyle=self._settings.readSettingAsString('minorLineStyle'),
                    linewidth=self._settings.readSettingAsString('minorLineWidth'), which='minor')

        ax.set_xlabel('covered days')
        ax.set_ylabel('filtered daily counts')
        if showValues:
            # TBD
            pass
        if self._settings.readSettingAsString('cursorEnabled') == 'true':
            cursor(hover=True)

        self._fig.set_tight_layout({"pad": 5.0})
        self._canvas = FigureCanvasQTAgg(self._fig)
        # avoids showing the original fig window
        plt.close('all')

    def zoom(self, horz: int, vert: int):
        self._fig.set_size_inches(horz, vert)
        self._canvas.draw()




"""
******************************************************************************

    Echoes Data Browser (Ebrow) is a data navigation and report generation
    tool for Echoes.
    Echoes is a RF spectrograph for SDR devices designed for meteor scatter.

    Both copyright (C) 2018-2023
    Giuseppe Massimo Bertani gm_bertani(a)yahoo.it

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html

*******************************************************************************

"""
import os
from PyQt5.QtGui import QColor, QPixmap
from PyQt5.QtCore import Qt, QRect, QSettings, QDate, QLocale
from .logprint import print

class Settings(QSettings):

    DBFS_RANGE_MIN = -150
    DBFS_RANGE_MAX = 20
    ZOOM_MIN = 0.5
    ZOOM_MAX = 10.0
    ZOOM_DEFAULT = 1.0
    AZIMUTH_DEFAULT = 0
    ELEVATION_DEFAULT = 45

    def __init__(self, iniFullPath):
        QSettings.__init__(self, iniFullPath, QSettings.IniFormat)
        self._locale = QLocale()
        print("Local country {}, language {}".format(QLocale.countryToString(self._locale.country()),
                                                               QLocale.languageToString(self._locale.language())))

        self._iniFilePath = iniFullPath
        self._general = ['geometry', 'lastDBfilePath', 'dateFrom', 'dateTo', 'tooltipDisabled', 'autosaving']

        self._screenshots = ['classFilter', 'currentDate', 'horizontalZoom', 'verticalZoom', 'linkedSliders', 'showGrid',
                             '3Dazimuth', '3Delevation', 'currentColormap']

        self._stats = ['classFilterStat', 'showValues', 'showGridStat', 'smoothPlots', 'subtractSporadicBackground',
                       'horizontalZoomStat', 'verticalZoomStat', 'linkedSlidersStat',
                       'currentColormapStat', 'RMOBfilePrefix',
                       'sporadicDates', 'sporadicBackgroundDaily', 'sporadicBackgroundByHour',
                       'sporadicBackgroundBy10min']

        self._plotStyles = ['fontFamily', 'fontSize', 'fontStyle','majorLineStyle', 'majorLineWidth', 'minorLineStyle',
                            'minorLineWidth', 'dataLineStyle', 'dataLineWidth',
                            'cursorEnabled',  'separator']

        self._siteInfos = ['stationName', 'owner', 'country', 'city', 'latitude', 'longitude', 'latitudeDeg',
                           'longitudeDeg', 'altitude', 'logoPath',
                           'antenna', 'antAzimuth', 'antElevation', 'preamplifier', 'receiver', 'frequencies',
                           'computer',
                           'email', 'notes']

        self._filters = ['ESDfilter', 'ESDfilterThreshold', 'SATfilter', 'SATfilterThreshold', 'CAR1filter',
                         'CAR1filterThreshold', 'CAR2filter', 'CAR2filterThreshold',
                         'underdenseMs', 'overdenseSec', 'acqActive']

        self._report = ['siteInfosExc', 'dailyExc', 'RMOBexc', 'hourlyExc', 'tenMinExc',
                        'chronoExc', 'expEvExc', 'stTabExc', 'stGrpExc', 'setupExc',
                        'classFilterReport', 'RMOBclient', 'preface']

        self._groups = {'General': self._general, 'Plotting styles': self._plotStyles, 'Site infos': self._siteInfos,
                        'Screenshots': self._screenshots, 'Stats': self._stats, 'Filters': self._filters,
                        'Report': self._report}

        self._colorMaps = ['echoes', 'colorgramme', 'gray', 'cividis', 'inferno', 'magma', 'plasma', 'viridis']

        d = dict()

        # default settings
        # warning: use 0/1 instead of False/True
        # general
        d['geometry'] = QRect(0, 0, 1024, 768)
        d['dateFrom'] = '2022-01-01'
        d['dateTo'] = '2022-10-31'
        d['lastDBfilePath'] = None
        d['tooltipDisabled'] = 0
        d['autosaving'] = 0

        # plot styles
        if os.name == 'nt':
            d['fontFamily'] = 'Times New Roman'
        else:
            d['fontFamily'] = 'LiberationSerif'

        d['fontSize'] = 6
        d['fontStyle'] = 'Normal'   # 'Italic Bold' is still possible
        d['dataLineStyle'] = 'solid'
        d['majorLineStyle'] = 'solid'
        d['minorLineStyle'] = 'solid'
        d['dataLineWidth'] = '1.0'
        d['majorLineWidth'] = '0.8'
        d['minorLineWidth'] = '0.5'
        d['cursorEnabled'] = 1
        d['separator'] = 'comma'

        # plotting colors
        d['colorDict'] = {'S': QColor('#ff0000'), 'N': QColor('#00ffff'), 'diff': QColor('#FF8C00'),
                          'avgDiff': QColor('#800000'), 'upperThr': QColor('#FFFF00'), 'lowerThr': QColor('#00FF00'),
                          'counts': QColor('#008080'), 'majorGrids': QColor('#696969'),
                          'minorGrids': QColor('#696969'), 'background': QColor('#FFFFFF')}

        # siteinfos
        d['stationName'] = ''
        d['owner'] = 'Somebody Somewhere'
        d['country'] = ''
        d['city'] = ''
        d['latitudeDeg'] = "0° 0\' 0\""
        d['longitudeDeg'] = "0° 0\' 0\""
        d['latitude'] = '0.0'
        d['longitude'] = '0.0'
        d['altitude'] = 0
        d['logoPath'] = ":/logo"
        d['antenna'] = ''
        d['antAzimuth'] = 0
        d['antElevation'] = 0
        d['frequencies'] = "0000.000.000"
        d['preamplifier'] = ''
        d['receiver'] = ''
        d['computer'] = ''
        d['email'] = ''
        d['notes'] = ''

        # screenshots
        d['currentDate'] = ''
        d['classFilter'] = 'OVER'
        d['horizontalZoom'] = self.ZOOM_DEFAULT
        d['verticalZoom'] =  self.ZOOM_DEFAULT
        d['linkedSliders'] = 0
        d['showGrid'] = 0
        d['3Dazimuth'] =  self.AZIMUTH_DEFAULT
        d['3Delevation'] =  self.ELEVATION_DEFAULT
        d['currentColormap'] = self._colorMaps[0] # default colormap for screenshots: echoes

        # stats
        d['classFilterStat'] = 'OVER'
        d['showGridStat'] = 0
        d['subtractSporadicBackground'] = 0
        d['showValues'] = 1
        d['smoothPlots'] = 1
        d['horizontalZoomStat'] = self.ZOOM_DEFAULT
        d['verticalZoomStat'] = self.ZOOM_DEFAULT
        d['linkedSlidersStat'] = 0
        d['currentColormapStat'] = self._colorMaps[1] # default colormap for statistics: colorgramme
        d['RMOBfilePrefix'] = 'UNKNOWN'
        d['sporadicDates'] = ''     # date intervals for sporadic backgroud calculation
        d['sporadicBackgroundDaily'] = ''
        d['sporadicBackgroundByHour'] = ''
        d['sporadicBackgroundBy10min'] = ''

        # filters
        d['ESDfilter'] = 1
        d['ESDfilterThreshold'] = 20.0
        d['SATfilter'] = 1
        d['SATfilterThreshold'] = 5.0
        d['CAR1filter'] = 1
        d['CAR1filterThreshold'] = 12.0
        d['CAR2filter'] = 1
        d['CAR2filterThreshold'] = 10.0
        d['underdenseMs'] = 200
        d['overdenseSec'] = 50
        d['acqActive'] = 0

        # report
        d['siteInfosExc'] = 0
        d['dailyExc'] = 0
        d['RMOBexc'] = 0
        d['hourlyExc'] = 0
        d['tenMinExc'] = 1
        d['chronoExc'] = 0
        d['expEvExc'] = 0
        d['stTabExc'] = 0
        d['stGrpExc'] = 0
        d['setupExc'] = 0
        d['classFilterReport'] = "OVER, UNDER"
        d['RMOBclient'] = "unknown"
        d['preface'] = ""

        self._defaults = d
        self._settings = d

    def decimalPoint(self):
        return self._locale.decimalPoint()

    def dataSeparator(self):
        sepIdx = self.readSettingAsString('separator')
        separators = {'space': ' ', 'tabulation': '\t', 'comma': ',', 'semicolon': ';', 'colon': ':', 'pipe': '|'}
        return separators[sepIdx]

    def writeSetting(self, key: str, value: any):
        self._settings[key] = value

    def readSettingAsObject(self, key: str):
        d = self._settings
        return self._settings[key]

    def readSettingAsString(self, key: str, default: str = '----'):
        d = self._settings
        s = str(self._settings[key])
        if len(s) == 0:
            s = default
        return s

    def readSettingAsInt(self, key: str):
        d = self._settings
        value = self._settings[key]
        return int(value)

    def readSettingAsBool(self, key: str):
        d = self._settings
        value = self._settings[key]
        if value == 'true':
            return True
        return False

    def readSettingAsFloat(self, key: str):
        d = self._settings
        value = self._settings[key]
        return float(value)

    def coverage(self):
        """
        @return: tuple from,to
        """
        d = self._settings
        return d['dateFrom'], d['dateTo']

    def load(self):
        for group in self._groups.keys():
            self.beginGroup(group)
            for key in self._groups[group]:
                try:
                    value = self.value(key, self._defaults[key])
                    # print("loading key={}, value={}, type is {}".format(key, value, type(value)))
                except KeyError:
                    value = 0
                self._settings[key] = value
            self.endGroup()

        self.beginGroup('Plotting  styles')
        colorNames = self._getDictValues(self._defaults['colorDict'], "Plotting colors")
        self._settings['colorDict'] = dict()
        for key in colorNames.keys():
            name = colorNames[key]
            color = QColor(name)
            self._settings['colorDict'][key] = color
        self.endGroup()

    def save(self):
        for group in self._groups.keys():
            self.beginGroup(group)
            for key in self._groups[group]:
                value = self._settings[key]
                # print("saving key={}, value={}, type is {}".format(key, value, type(value)))
                self.setValue(key, value)
            self.endGroup()

        colorNames = dict()
        for key in self._settings['colorDict'].keys():
            name = self._settings['colorDict'][key].name()
            colorNames[key] = name
        self.beginGroup('Plotting styles')
        self._setDictValues(colorNames, "Plotting colors")
        self.endGroup()
        self.sync()

    def _setDictValues(self, values: dict, groupName: str):
        self.beginGroup(groupName)
        for key in values.keys():
            self.setValue(key, values[key])
        self.endGroup()

    def _getDictValues(self, defaults: dict, groupName: str):
        self.beginGroup(groupName)
        values = dict()
        for key in defaults.keys():
            values[key] = self.value(key, defaults[key])
        self.endGroup()
        return values



"""
******************************************************************************

    Echoes Data Browser (Ebrow) is a data navigation and report generation
    tool for Echoes.
    Echoes is a RF spectrograph for SDR devices designed for meteor scatter
    Both copyright (C) 2018-2023
    Giuseppe Massimo Bertani gm_bertani(a)yahoo.it

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html

*******************************************************************************

"""

import pandas as pd
import numpy as np
import matplotlib as mp
import matplotlib.pyplot as plt
from PyQt5.QtCore import QDate
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from mplcursors import cursor

from .settings import Settings
from .basegraph import BaseGraph
from .logprint import print

mp.use('Qt5Agg')


class Heatmap(BaseGraph):
    def __init__(self, dataFrame: pd.DataFrame, settings: Settings, inchWidth: float, inchHeight: float, cmap: list,
                 title: str, res: str = 'hour', showValues: bool = False, showGrid: bool = False, subtractBackground: bool = False):
        BaseGraph.__init__(self, settings)

        xlabel = "covered days"
        ylabel = "hours of a day"
        colors = self._settings.readSettingAsObject('colorDict')

        if res == "hour":  # len(dataFrame.columns) == 24:
            if subtractBackground:
                title += "\nafter sporadic background subtraction"
            # 24 columns, one for each hour count from 00h to 23h
            df = dataFrame
            days = len(df.index)
            # calculate full-scale value (red color) taking the maximum in all the counts columns
            countsOnlyDf = df.iloc[0:, 0:]
            self._fullScale = countsOnlyDf.max(axis=1).max()

            # rows header (hours)
            rowsHdr = list()
            # rowsHdr.append("Day / Hour")
            for hour in range(0, 24):
                s = "{:02d}h".format(hour)
                rowsHdr.append(s)

            # columns header (days)
            allColumns = list()
            colsHdr = list()
            for dayCount in range(0, days):
                isoDate = str(df.index[dayCount])
                if len(isoDate.split('-')) == 3:
                    date = QDate.fromString(isoDate, "yyyy-MM-dd")
                    s = date.day()
                else:
                    s = isoDate
                colsHdr.append(s)

                # days columns values
                column = list()
                for hour in range(0, 24):
                    value = int(df.iloc[dayCount, hour])
                    column.append(value)
                allColumns.append(column)

        elif res == '10m':  # len(dataFrame.columns) == 144:
            if subtractBackground:
                title += "\nafter sporadic background subtraction"
            # 144 columns, one for each 10min count from 00h.00m to 23h50m
            df = dataFrame
            days = len(df.index)

            # calculate full-scale value (red color) taking the maximum in all the counts columns
            # (the date column must be excluded for a correct calculation)

            countsOnlyDf = df.iloc[0:, 0:]
            self._fullScale = countsOnlyDf.max(axis=1).max()
            # rows header (hours)
            rowsHdr = list()
            # rows header (10minute intervals)
            interval = 0
            for hour in range(0, 24):
                for minute in range(0, 60, 10):
                    s = "{:02d}h {:02d}m".format(hour, minute)
                    interval += 1
                    rowsHdr.append(s)
            allColumns = list()

            # columns header (days)
            colsHdr = list()
            for dayCount in range(0, days):
                isoDate = str(df.index[dayCount])
                if len(isoDate.split('-')) == 3:
                    date = QDate.fromString(isoDate, "yyyy-MM-dd")
                    s = date.day()
                else:
                    s = isoDate
                colsHdr.append(s)

                # days columns value
                interval = 0
                column = list()
                for hour in range(0, 24):
                    for minute in range(0, 60, 10):
                        value = int(df.iloc[dayCount, interval])
                        column.append(value)
                        interval += 1

                allColumns.append(column)

        elif res == 'day':  # 1 <= len(dataFrame.columns) <= 8:
            # up to 7 columns, one for each event class filtered + totals
            xlabel = "events classification"
            ylabel = "days covered"
            if subtractBackground:
                title += "\nafter sporadic background subtraction"
            # discard the daily totals, currently on rows
            # df = dataFrame.drop(labels=['Total'], axis=0)
            df = dataFrame

            # calculate full-scale value (red color) taking the maximum in all the counts columns
            countsOnlyDf = df.iloc[0:, 0:]
            self._fullScale = countsOnlyDf.max(axis=1).max()

            # rows header (days)
            rowsHdr = list()
            for day in df.index:
                s = "{}".format(day)
                rowsHdr.append(s)

            allColumns = list()

            # columns header (classes)
            colsHdr = []
            colsList = list(dataFrame.columns)
            for col in colsList:
                col = col.replace(' ', '\n')
                colsHdr.append(col)

            # columns counts per class by day
            for flt in range(0, len(colsHdr)):
                column = list()
                for dayCount in range(0, len(rowsHdr)):
                    print("flt={}, dayCount={}".format(flt, dayCount))
                    count = int(df.iloc[dayCount, flt])
                    column.append(count)
                allColumns.append(column)
        else:
            return

        data = np.array(allColumns)
        data = np.transpose(data)

        backColor = colors['background'].name()
        plt.figure(figsize=(inchWidth, inchHeight))
        self._fig, ax = plt.subplots(1, facecolor=backColor)
        ax.set_facecolor(backColor)

        im = ax.imshow(data, cmap=cmap, aspect='auto')

        # Show all ticks and label them with the respective list entries
        if showGrid:
            ax.spines[:].set_visible(True)
            ax.set_xticks(np.arange(len(colsHdr)), labels=colsHdr)  # , minor=True)
            ax.set_yticks(np.arange(len(rowsHdr)), labels=rowsHdr)  # , minor=True)
            ax.grid(which="minor", color=colors['minorGrids'].name(),
                    linestyle=self._settings.readSettingAsString('minorLineStyle'),
                    linewidth=self._settings.readSettingAsString('minorLineWidth'))
            ax.tick_params(which="minor", bottom=False, left=False)
        else:
            ax.spines[:].set_visible(False)
            ax.set_xticks(np.arange(len(colsHdr)), labels=colsHdr)
            ax.set_yticks(np.arange(len(rowsHdr)), labels=rowsHdr)

        ax.xaxis.set_ticks_position('top')
        ax.xaxis.set_label_position('top')
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.tick_params(axis='x', which='both', labelrotation=60)
        ax.set_title(title, loc='left')
        plt.setp(ax.get_xticklabels(), ha="center")
        valFmt = mp.ticker.StrMethodFormatter('{x:.0f}')
        ax.tick_params(axis='x', which='both', labelrotation=60)
        if showValues:
            # Loop over data dimensions and create text annotations.
            for i in range(len(rowsHdr)):
                for j in range(len(colsHdr)):
                    if data[i, j] > (self._fullScale / 4):
                        ax.text(j, i, valFmt(data[i, j]), ha="center", va="center", color="black")
                    else:
                        ax.text(j, i, valFmt(data[i, j]), ha="center", va="center", color="white")

        norm = mp.colors.Normalize(vmin=0, vmax=int(self._fullScale))
        self._fig.colorbar(im, drawedges=showGrid, norm=norm, cmap=cmap)

        if self._settings.readSettingAsString('cursorEnabled') == 'true':
            cursor(hover=True)

        self._fig.set_tight_layout({"pad": 3.0})
        self._canvas = FigureCanvasQTAgg(self._fig)
        # avoids showing the original fig window
        plt.close('all')

    def zoom(self, horz: int, vert: int):
        self._fig.set_size_inches(horz, vert)
        self._canvas.draw()


"""
******************************************************************************

    Echoes Data Browser (Ebrow) is a data navigation and report generation
    tool for Echoes.
    Echoes is a RF spectrograph for SDR devices designed for meteor scatter
    Both copyright (C) 2018-2023
    Giuseppe Massimo Bertani gm_bertani(a)yahoo.it

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html

*******************************************************************************

"""
import os

import pandas as pd
from .pandasmodel import PandasModel
from .logprint import print


class RTSconfig:
    RTS_DEVICES = 0
    RTS_FFT = 1
    RTS_OUTPUT = 2
    RTS_STORAGE = 3
    RTS_PREFS = 4
    RTS_TOTAL = 5

    def __init__(self, parent, ui, settings):
        self._ui = ui
        self._parent = parent
        self._settings = settings
        self._configRevision = 0
        self._dataSource = self._parent.dataSource
        self._currentDf = None
        self._ui.lbID_2.setText('ID# ' + str(self._parent.currentID))
        self._ui.lbDaily_2.setText('Daily# ' + str(self._parent.currentDailyNr))
        self._ui.lbDay.setText(self._parent.currentDate)
        self._ui.twRTS.currentChanged.connect(self.updateTabRTSconfig)
        self._ui.pbCfgTabExp.clicked.connect(self._exportPressed)

    def updateTabRTSconfig(self):
        self._parent.busy(True)
        self._dataSource = self._parent.dataSource
        self._ui.lbID_2.setText('ID# ' + str(self._parent.currentID))
        self._ui.lbDaily_2.setText('Daily# ' + str(self._parent.currentDailyNr))
        self._ui.lbDay.setText(self._parent.currentDate)
        self._configRevision = self._dataSource.getCfgRevisionFromID(self._parent.currentID)
        self._ui.lbCfgRev.setNum(self._configRevision)
        if self._ui.twMain.currentIndex() == self._parent.TWMAIN_RTS:
            if self._ui.twRTS.currentIndex() == self.RTS_DEVICES:
                self._displayCfgDevices()
            if self._ui.twRTS.currentIndex() == self.RTS_FFT:
                self._displayCfgFFT()
            if self._ui.twRTS.currentIndex() == self.RTS_OUTPUT:
                self._displayCfgOutput()
            if self._ui.twRTS.currentIndex() == self.RTS_STORAGE:
                self._displayCfgStorage()
            if self._ui.twRTS.currentIndex() == self.RTS_PREFS:
                self._displayCfgPrefs()
        self._parent.busy(False)

    def _displayCfgDevices(self):
        self._currentDf = self._dataSource.loadTableConfig('cfg_devices', self._configRevision)
        model = PandasModel(self._currentDf)
        self._ui.tvCfgDevice.setModel(model)

    def _displayCfgFFT(self):
        self._currentDf = self._dataSource.loadTableConfig('cfg_fft', self._configRevision)
        model = PandasModel(self._currentDf)
        self._ui.tvCfgFFT.setModel(model)

    def _displayCfgOutput(self):
        self._currentDf = self._dataSource.loadTableConfig('cfg_output', self._configRevision)
        model = PandasModel(self._currentDf)
        self._ui.tvCfgOutput.setModel(model)

    def _displayCfgFilters(self):
        self._currentDf = self._dataSource.loadTableConfig('cfg_filters', self._configRevision)
        model = PandasModel(self._currentDf)
        self._ui.tvCfgFilters.setModel(model)

    def _displayCfgStorage(self):
        self._currentDf = self._dataSource.loadTableConfig('cfg_storage', self._configRevision)
        model = PandasModel(self._currentDf)
        self._ui.tvCfgStorage.setModel(model)

    def _displayCfgPrefs(self):
        self._currentDf = self._dataSource.loadTableConfig('cfg_prefs', self._configRevision)
        model = PandasModel(self._currentDf)
        self._ui.tvCfgPrefs.setModel(model)

    def _exportPressed(self, checked):
        self._parent.busy(True)
        # the displayed table is exported as csv
        os.chdir(self._parent.exportDir)
        tableNames = ['Device', 'FFT', 'Output', 'Filters', 'Storage', 'Preferences']
        title = tableNames[self._ui.twRTS.currentIndex()]
        self._currentDf.style.set_caption(title)
        title = title.lower()
        csvName = 'RTS_' + title + '_config.csv'
        self._currentDf.to_csv(csvName, index=True, sep=self._settings.dataSeparator(),
                               decimal=self._settings.decimalPoint())
        self._parent.updateStatusBar("Exported  {}".format(csvName))
        self._ui.lbCfgTabFilename.setText(csvName)
        os.chdir(self._parent.workingDir)
        self._parent.busy(False)

    def getAllTables(self, rev):
        tables = ['cfg_devices', 'cfg_fft', 'cfg_output', 'cfg_storage', 'cfg_prefs']
        rtsDf = list()
        for table in tables:
            df = self._dataSource.loadTableConfig(table, rev)
            rtsDf.append(df)
        return rtsDf



"""
******************************************************************************

    Echoes Data Browser (Ebrow) is a data navigation and report generation
    tool for Echoes.
    Echoes is a RF spectrograph for SDR devices designed for meteor scatter
    Both copyright (C) 2018-2023
    Giuseppe Massimo Bertani gm_bertani(a)yahoo.it

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html

*******************************************************************************

"""

import pandas as pd
import numpy as np
import matplotlib as mp
import matplotlib.pyplot as plt
import matplotlib.dates as md
from matplotlib.ticker import MultipleLocator
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg

from .settings import Settings
from .basegraph import BaseGraph
from .logprint import print

mp.use('Qt5Agg')


class Bargraph(BaseGraph):
    def __init__(self, series: pd.Series, settings: Settings, inchWidth: float, inchHeight: float, title: str, res: str = 'hour',
                 showValues: bool = False, showGrid: bool = False, subtractBackground = False):
        BaseGraph.__init__(self, settings)
        self._series = series
        x = self._series.index
        y = self._series.values

        colors = self._settings.readSettingAsObject('colorDict')

        backColor = colors['background'].name()
        plt.figure(figsize=(inchWidth, inchHeight))
        self._fig, ax = plt.subplots(1, facecolor=backColor)
        ax.set_facecolor(backColor)

        xdt = np.asarray(x, dtype='datetime64[s]')
        ax.xaxis.set_major_locator(MultipleLocator(1))
        myFmt = md.DateFormatter('%Y-%m-%d')
        ax.xaxis.set_major_formatter(myFmt)
        ax.tick_params(axis='x', which='both', labelrotation=60)

        if res == '10m':
            if subtractBackground:
                title += "\nafter sporadic background subtraction"
            ax.bar(xdt, y, width=1 / 144, color=colors['counts'].name(), edgecolor=backColor, align="edge")
            locator = MultipleLocator(6 / 144)
            locator.MAXTICKS = 10000
            ax.xaxis.set_minor_locator(locator)
            myFmt = md.DateFormatter('%Hh')
            ax.xaxis.set_minor_formatter(myFmt)

        elif res == 'day':
            if subtractBackground:
                title += "\nafter sporadic background subtraction"
            # ax.bar(xdt, y,  width=0.1, color=colors['counts'].name(), edgecolor="white", align="edge")
            ax.bar(xdt, y, color=colors['counts'].name(), edgecolor="white", align="edge")

        elif res == 'hour':
            if subtractBackground:
                title += "\nafter sporadic background subtraction"
            ax.bar(xdt, y, width=1 / 24, color=colors['counts'].name(), edgecolor="white", align="edge")
            ax.xaxis.set_minor_locator(MultipleLocator(6 / 24))
            myFmt = md.DateFormatter('%Hh')
            ax.xaxis.set_minor_formatter(myFmt)

        if showValues:
            # TBD
            pass

        if showGrid:
            # TBD
            pass

        ax.set_title(title, loc='left')
        self._fig.set_tight_layout({"pad": 5.0})
        self._canvas = FigureCanvasQTAgg(self._fig)
        # avoids showing the original fig window
        plt.close('all')

    def zoom(self, horz: int, vert: int):
        self._fig.set_size_inches(horz, vert)
        self._canvas.draw()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    # Here is the module name.
    name="ebrow",

    # version of the module
    # !!! GMB WARNING -  IT MUST MATCH THE VERSION IN pyproject.toml and maiwindow.py !!!!!!
    version="0.1.70",

    # Name of Author
    author="Bertani Giuseppe Massimo",

    # your Email address
    author_email="gm_bertani@yahoo.it",

    # #Small Description about module
    description="Echoes Data Browser (Ebrow) is a data navigation and report generation tool for Echoes.",

    long_description=long_description,

    # Specifying that we are using markdown file for description
    long_description_content_type="text/markdown",

    # Any link to reach this module, ***if*** you have any webpage or github profile
    url="https://www.gabb.it/echoes/",
    packages=setuptools.find_packages(),

    # if module has dependencies i.e. if your package rely on other package at pypi.org
    # then you must add there, in order to download every requirement of package

    install_requires = [ # Optional
        "matplotlib>=3.7.1",
        "numpy>=1.24.3",
        "pandas>=2.0.0",
        "PyQt5>=5.15.7",
        "setuptools>=61.0",
        "mplcursors>=0.5.2",
        "python-dateutil>=2.8.2",
        "pyqtspinner>=2.0.0",
        "openpyxl>=3.1.2",
        "jinja2>=3.1.2",
	    "scipy>=1.11"
    ],

    license="GPLv3",

    # classifiers like program is suitable for python3, just leave as it is.
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Opebrating System :: OS Independent",
    ],
)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'dateintervaldialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_DateIntervalDialog(object):
    def setupUi(self, DateIntervalDialog):
        DateIntervalDialog.setObjectName("DateIntervalDialog")
        DateIntervalDialog.setWindowModality(QtCore.Qt.WindowModal)
        DateIntervalDialog.resize(722, 350)
        DateIntervalDialog.setMinimumSize(QtCore.QSize(700, 350))
        DateIntervalDialog.setStyleSheet("/******************************************************************************\n"
"\n"
"    Echoes Data Browser (Ebrow) is a data navigation and report generation\n"
"    tool for Echoes.\n"
"    Echoes is a RF spectrograph for SDR devices designed for meteor scatter.\n"
"    \n"
"    Both copyright (C) 2018-2023\n"
"    Giuseppe Massimo Bertani gm_bertani(a)yahoo.it\n"
"\n"
"    This program is free software: you can redistribute it and/or modify\n"
"    it under the terms of the GNU General Public License as published by\n"
"    the Free Software Foundation, version 3 of the License.\n"
"\n"
"    This program is distributed in the hope that it will be useful,\n"
"    but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"    GNU General Public License for more details.\n"
"\n"
"    You should have received a copy of the GNU General Public License\n"
"    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html\n"
"\n"
"\n"
"********************************************************************************/\n"
"\n"
"/* default colors and font for all the widgets */\n"
"QWindow {\n"
"\n"
"    border-radius: 6px; \n"
"    border: 1px solid rgb(127, 127, 127);\n"
"} \n"
"\n"
"QMessageBox {\n"
"\n"
"    border-radius: 6px; \n"
"    border: 0px solid rgb(127, 127, 127);\n"
"} \n"
"\n"
"QWidget, QWidget * {\n"
"    background-color: rgb(40, 40, 40);\n"
"    font: 87 8pt \"Gauge\";\n"
"    color:rgb(255, 255, 127);\n"
"    border-radius: 6px; \n"
"    border: 0px solid rgb(40, 40, 40); /* invisible border */\n"
"} \n"
"\n"
"QWidget::disabled {\n"
"    background-color: rgb(40, 40, 40);\n"
"    font: 8pt \"Gauge\";\n"
"    color:rgb(127, 127, 127);\n"
"    border-radius: 6px;\n"
"    border: 1px solid rgb(127, 127, 127); /* visible border */\n"
"}\n"
"\n"
"\n"
"QTabWidget, QGroupBox {\n"
"    background-color: rgb(40, 40, 40);\n"
"    color:rgb(255, 255, 127);\n"
"    border-radius: 6px; \n"
"    border: 1px solid rgb(148, 148, 0);  /* yellow border */\n"
"    margin: 3px;\n"
"}\n"
"\n"
"QListWidget, QListView, QTableView, QTableWidget, QScrollArea {\n"
"    border: 1px solid rgb(148, 148, 0);  /* yellow border */\n"
"     show-decoration-selected: 1;          /* make the selection span the entire width of the view */\n"
"\n"
"}\n"
"\n"
"QCheckBox:disabled {\n"
"    border: none;\n"
"}\n"
"\n"
"QCheckBox:enabled {\n"
"    border: yellow;\n"
"}\n"
"\n"
"QLabel {\n"
"    margin: 0 0;\n"
"}\n"
"\n"
"QLabel:disabled {\n"
"    border: none;\n"
"}\n"
"\n"
"\n"
"QListWidget::item:selected:active {\n"
"    color: black;\n"
"    font-weight: bold;\n"
"    background-color: yellow;\n"
"}\n"
"\n"
"QListWidget::item:selected:!active  {\n"
"    color:rgb(255, 255, 127);\n"
"    background-color: green; /*rgb(40, 40, 40);*/\n"
"    font-weight: normal;\n"
"}\n"
"\n"
"QListWidget::item:normal  {\n"
"    color:rgb(255, 255, 127);\n"
"    background-color: rgb(40, 40, 40);\n"
"    font-weight: normal;\n"
"}\n"
"\n"
"\n"
"QTableView, QTableWidget {\n"
"    alternate-background-color: rgb(60, 60, 60);\n"
"    color:rgb(255, 255, 127);\n"
"    gridline-color:  rgb(156, 156, 156);\n"
"    border-radius: 4px;\n"
"}\n"
"\n"
"QGroupBox, QLabel {\n"
"    margin-top: 10px;\n"
"    \n"
"}\n"
"\n"
"QTabBar::tab {\n"
"    background: rgb(0, 96, 100);\n"
"    min-width: 120px;\n"
"}\n"
"\n"
"QTabBar::tab:disabled {\n"
"    color: grey;\n"
"    background: rgb(34, 34, 51);\n"
"}\n"
"\n"
"QTabBar::tab:enabled {\n"
"    background: rgb(34, 34, 51);\n"
"}\n"
"\n"
"QTabBar::tab:selected {\n"
"    background: rgb(0, 96, 250);\n"
"}\n"
"\n"
"QHeaderView::section {\n"
"     background-color:  rgb(230, 230, 230);\n"
"     color: rgb(20, 20, 20);\n"
"} \n"
"\n"
"QGroupBox::title {\n"
"    subcontrol-origin: margin;\n"
"    subcontrol-position: top center; /* position at the top center */\n"
"    padding: 0 3px;\n"
"}\n"
"\n"
"/* personalization for input widgets */\n"
"QLineEdit, QPlainTextEdit, QDateTimeEdit, QAbstractSpinBox, QComboBox, QLCDNumber {\n"
"    background-color: rgb(10, 10, 10); \n"
"    color: rgb(0, 255, 0);\n"
"    selection-background-color: rgb(85, 85, 255);\n"
"    selection-color: rgb(0, 255, 255);\n"
"    border-radius: 6px; \n"
"\n"
"}\n"
"\n"
"/* tooltips cannot be tuned by QtDesigner, it must be done explicitly by application */\n"
"QToolTip {\n"
"    background: blue;\n"
"    color: yellow;\n"
"    border: 1px;\n"
"    font: 8pt \"Arial\";\n"
"}\n"
"\n"
"QPushButton {\n"
"    background-color:  rgb(0, 96, 100);\n"
"    min-height: 20px;\n"
"    min-width: 60px;\n"
"    text-align: center;\n"
"}\n"
"\n"
"QToolButton {\n"
"    background-color:  rgb(0, 96, 100);\n"
"    min-height: 24px;\n"
"    min-width: 32px\n"
"}\n"
"\n"
"QPushButton::pressed {\n"
"    background-color:  red;\n"
"    min-height: 24px;\n"
"    min-width: 32px\n"
"}\n"
"\n"
"QToolButton::pressed {\n"
"    background-color:  red;\n"
"    min-height: 24px;\n"
"    min-width: 32px\n"
"}\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"/*\n"
"QSlider::groove:vertical {\n"
"    background: black;\n"
"    position: absolute;\n"
"    left: 2px; right: 2px;\n"
"}\n"
"\n"
"\n"
"\n"
"QSlider::handle:vertical {\n"
"    height: 20px;\n"
"    width: 30px;\n"
"    background: qlineargradient(x1:0, y1:0, x2:3, y2:1, stop:0 grey, stop:1 white);\n"
"    margin: 0 -20px; \n"
"    border: 2px  qlineargradient(x1:0, y1:0, x2:3, y2:1, stop:0 white, stop:1 grey);\n"
"    border-radius: 4px;\n"
"}\n"
"\n"
"QSlider::groove:horizontal {\n"
"    background: black;\n"
"    position: absolute;\n"
"    top: 1px; bottom: 1px;\n"
"}\n"
"*/\n"
"QSlider::handle:horizontal {\n"
"    height: 30px;\n"
"    width: 20px;\n"
"    background: qlineargradient(x1:0, y1:0, x2:3, y2:1, stop:0 grey, stop:1 white);\n"
"    margin: 0 -20px; /* expand outside the groove */\n"
"    border: 2px  qlineargradient(x1:0, y1:0, x2:3, y2:1,  stop:0 white, stop:1 grey);\n"
"    border-radius: 4px;\n"
"}\n"
"")
        DateIntervalDialog.setSizeGripEnabled(True)
        DateIntervalDialog.setModal(True)
        self.verticalLayout = QtWidgets.QVBoxLayout(DateIntervalDialog)
        self.verticalLayout.setSpacing(12)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.cwFrom = QtWidgets.QCalendarWidget(DateIntervalDialog)
        self.cwFrom.setObjectName("cwFrom")
        self.horizontalLayout_2.addWidget(self.cwFrom)
        self.cwTo = QtWidgets.QCalendarWidget(DateIntervalDialog)
        self.cwTo.setObjectName("cwTo")
        self.horizontalLayout_2.addWidget(self.cwTo)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.pbOk = QtWidgets.QPushButton(DateIntervalDialog)
        self.pbOk.setMinimumSize(QtCore.QSize(60, 20))
        self.pbOk.setMaximumSize(QtCore.QSize(200, 32))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("resources/icons/mail-filters-apply.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pbOk.setIcon(icon)
        self.pbOk.setObjectName("pbOk")
        self.horizontalLayout.addWidget(self.pbOk)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(DateIntervalDialog)
        QtCore.QMetaObject.connectSlotsByName(DateIntervalDialog)

    def retranslateUi(self, DateIntervalDialog):
        _translate = QtCore.QCoreApplication.translate
        DateIntervalDialog.setWindowTitle(_translate("DateIntervalDialog", "Dialog"))
        self.pbOk.setText(_translate("DateIntervalDialog", "&Ok"))



"""
******************************************************************************

    Echoes Data Browser (Ebrow) is a data navigation and report generation
    tool for Echoes.
    Echoes is a RF spectrograph for SDR devices designed for meteor scatter
    Both copyright (C) 2018-2023
    Giuseppe Massimo Bertani gm_bertani(a)yahoo.it

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html

*******************************************************************************

"""

import pandas as pd
import matplotlib as mp
import matplotlib.pyplot as plt
import matplotlib.dates as md
from dateutil.rrule import SECONDLY, MINUTELY
from mplcursors import cursor
from matplotlib.ticker import MaxNLocator
from matplotlib.dates import AutoDateLocator
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg

from .settings import Settings
from .basegraph import BaseGraph
from .logprint import print

mp.use('Qt5Agg')


class PowerPlot(BaseGraph):
    def __init__(self, df: pd.DataFrame, name: str, settings: Settings, inchWidth: float, inchHeight: float,
                 xLocBaseSecs: int = 5, yMinTicks: int = 10, showGrid: bool = False):
        BaseGraph.__init__(self, settings)

        colors = self._settings.readSettingAsObject('colorDict')
        backColor = colors['background'].name()
        plt.figure(figsize=(inchWidth, inchHeight))
        self._fig, ax = plt.subplots(1, facecolor=backColor)
        ax.set_facecolor(backColor)

        xLocator = AutoDateLocator()
        if xLocBaseSecs > 120:
            xLocBaseMins = xLocBaseSecs / 60
            xLocator.intervald[MINUTELY] = [xLocBaseMins]
        else:
            xLocator.intervald[SECONDLY] = [xLocBaseSecs]

        ax.xaxis.set_major_locator(xLocator)

        yLocator = MaxNLocator(min_n_ticks=yMinTicks)
        ax.yaxis.set_major_locator(yLocator)

        myFmt = md.DateFormatter('%H:%M:%S')
        ax.xaxis.set_major_formatter(myFmt)
        title = "Power profile from data file " + name
        self._min = df['S'].min()
        self._max = df['S'].max()
        ax.plot(df.index, df['N'], linestyle=self._settings.readSettingAsString('dataLineStyle'),
                linewidth=self._settings.readSettingAsString('dataLineWidth'), color=colors['N'].name(), label='N')
        ax.plot(df.index, df['S'], linestyle=self._settings.readSettingAsString('dataLineStyle'),
                linewidth=self._settings.readSettingAsString('dataLineWidth'), color=colors['S'].name(), label='S')
        ax.plot(df.index, df['diff'], linestyle=self._settings.readSettingAsString('dataLineStyle'),
                linewidth=self._settings.readSettingAsString('dataLineWidth'), color=colors['diff'].name(), label='S-N')
        ax.set_title(title, loc='left')
        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        ax.tick_params(axis='x', which='both', labelrotation=60)
        if showGrid:
            ax.grid(color=colors['majorGrids'].name(), linewidth=self._settings.readSettingAsString('majorLineWidth'),
                    linestyle=self._settings.readSettingAsString('majorLineStyle'), which='major')
        ax.set_xlabel('time of day')
        ax.set_ylabel('total power [dBfs]')
        if self._settings.readSettingAsString('cursorEnabled') == 'true':
            cursor(hover=True)

        self._df = df

        self._fig.set_tight_layout({"pad": 5.0})
        self._canvas = FigureCanvasQTAgg(self._fig)
        # avoids showing the original fig window
        plt.close('all')

    def savePlotDataToDisk(self, fileName):
        self._df.to_csv(fileName, sep=self._settings.dataSeparator())

    def zoom(self, horz: int, vert: int):
        self._fig.set_size_inches(horz, vert)
        # self._ax3d.set_xlim3d(-horz/2, horz/2)
        # self._ax3d.set_ylim3d(-vert / 2, vert / 2)
        self._canvas.draw()
        # plt.pause(0.01)

    def getMinMax(self):
        return [self._min, self._max]


"""
******************************************************************************

    Echoes Data Browser (Ebrow) is a data navigation and report generation
    tool for Echoes.
    Echoes is a RF spectrograph for SDR devices designed for meteor scatter
    Both copyright (C) 2018-2023
    Giuseppe Massimo Bertani gm_bertani(a)yahoo.it

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html

*******************************************************************************

"""

import pandas as pd
import numpy as np
import matplotlib as mp
import matplotlib.pyplot as plt
import matplotlib.dates as md
from matplotlib.ticker import MultipleLocator
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg

from .settings import Settings
from .basegraph import BaseGraph
from .logprint import print
mp.use('Qt5Agg')


class BargraphRMOB(BaseGraph):
    def __init__(self, series: pd.Series, settings: Settings, inchWidth: float, inchHeight: float, cmap: list, fullScale: int ):
        BaseGraph.__init__(self, settings, 'RMOB')
        self._series = series
        x = self._series.index
        y = self._series.values

        colors = self._settings.readSettingAsObject('colorDict')
        plt.rc('axes', linewidth=2)

        plt.figure(figsize=(inchWidth, inchHeight))

        self._fig, ax = plt.subplots(1, facecolor='#ffffff')
        ax.set_facecolor('#ffffff')
        # ax.set_aspect(0.008)
        xdt = np.asarray(x, dtype='datetime64[s]')
        ax.xaxis.set_major_locator(MultipleLocator(1))
        myFmt = md.DateFormatter('%Y-%m-%d')
        ax.xaxis.set_major_formatter(myFmt)
        ax.tick_params(axis='x', which='both', labelrotation=0)
        # ax.bar(xdt, y, width=1 / 30, color=colors['counts'].name(), edgecolor="white", align="edge", )
        # generate colorlist referred to maximum monthly value as fullscale, the same of the heatmap
        barColors = list()
        ratio = len(cmap.colors) / fullScale
        for value in y:
            idx = int(value * ratio)
            if idx >= len(cmap.colors):
                idx = len(cmap.colors)-1
            barColors.append(cmap.colors[idx])

        ax.bar(xdt, y, width=1 / 30, color=barColors, edgecolor="white", align="edge")
        ax.yaxis.set_ticklabels([])
        ax.set_yticks([])
        ax.xaxis.set_major_locator(MultipleLocator(4 / 24))
        myFmt = md.DateFormatter('%Hh')
        ax.xaxis.set_major_formatter(myFmt)
        self._maxVal = self._series.values.max(initial=0)
        ax.annotate(str(  self._maxVal), xy=(-0.1,   self._maxVal - 4.0), xycoords=("axes fraction", "data"))
        ax.axhline(y=  self._maxVal, color=colors['minorGrids'].name(), linestyle='--')
        self._fig.set_tight_layout({"w_pad": 1.0, "h_pad": 0.0})
        self._canvas = FigureCanvasQTAgg(self._fig)
        # avoids showing the original fig window
        plt.close('all')

    def maxVal(self):
        return self._maxVal
